# Copyright (c) Acconeer AB, 2022
# All rights reserved

from . import platforms as _  # noqa: F401
from .base import PlatformInstall, SetupStep, utils
from .cli import SetupArgumentParser, prompts


def main():
    args = SetupArgumentParser().parse_args()

    selected_platform = args.platform or prompts.get_selection_from_user(
        "Platforms:", PlatformInstall.platforms()
    )

    mby_setupper = PlatformInstall.from_key(str(selected_platform))
    if selected_platform is None or mby_setupper is None:
        exit()

    setupper: SetupStep = utils.WithDescription(
        f"==== Setting up {selected_platform!r} will do the following:\n",
        mby_setupper,
    )

    if setupper is None:
        print(f"Platform {selected_platform!r} is not supported.")
        exit(1)

    if not args.silent:
        setupper.report()

        if not prompts.yn_prompt("Proceed?"):
            return

    success = setupper.run()
    if success:
        print("All done!")
    else:
        print("Something went wrong.")


if __name__ == "__main__":
    try:
        main()
    except RuntimeError as re:
        print(*re.args)
    except KeyboardInterrupt:
        pass



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import platform
import webbrowser
from typing import Optional

import pyperclip
import qtawesome as qta

from PySide6 import QtCore, QtGui
from PySide6.QtWidgets import (
    QComboBox,
    QFrame,
    QHBoxLayout,
    QLabel,
    QMessageBox,
    QPushButton,
    QWidget,
)

from acconeer.exptool.app.new._enums import ConnectionInterface
from acconeer.exptool.app.new._exceptions import HandledException
from acconeer.exptool.app.new.app_model import AppModel


BUTTON_ICON_COLOR = "#0081db"


class ExceptionWidget(QMessageBox):
    def __init__(
        self,
        parent: QWidget,
        *,
        exc: Exception,
        traceback_str: Optional[str] = None,
        title: str = "Error",
    ) -> None:
        super().__init__(parent)

        self.setIcon(QMessageBox.Warning)
        self.setStandardButtons(QMessageBox.Ok)

        self.setWindowTitle(title)
        self.setText(str(exc))

        try:
            raise exc
        except HandledException:
            pass
        except Exception:
            self.setInformativeText("<b>Unhandled error - please file a bug</b>")

        if traceback_str:
            self.setDetailedText(traceback_str)
            copy_button = QPushButton(self)
            copy_button.setText("Copy details")
            self.addButton(copy_button, QMessageBox.ButtonRole.ActionRole)
            copy_button.clicked.disconnect()
            copy_button.clicked.connect(self._on_copy_clicked)

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        super().resizeEvent(event)
        self.setFixedWidth(500)

    def _on_copy_clicked(self) -> None:
        detailed_text = self.detailedText()
        if detailed_text:
            pyperclip.copy(detailed_text)

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        self.close()


class VerticalSeparator(QWidget):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setLayout(QHBoxLayout(self))
        self.layout().setContentsMargins(5, 5, 5, 5)

        frame = QFrame(self)
        frame.setFrameShape(QFrame.VLine)
        self.layout().addWidget(frame)


class SerialPortComboBox(QComboBox):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)
        self.app_model = app_model
        app_model.sig_notify.connect(self._on_app_model_update)
        self.currentTextChanged.connect(self._on_change)

    def _on_change(self) -> None:
        self.app_model.set_serial_connection_port(self.currentData())

    def _on_app_model_update(self, app_model: AppModel) -> None:
        tagged_ports = app_model.available_tagged_ports

        ports = [port for port, _ in tagged_ports]
        view_ports = [self.itemData(i) for i in range(self.count())]

        with QtCore.QSignalBlocker(self):
            if ports != view_ports:
                self.clear()
                for port, tag in tagged_ports:
                    label = port if tag is None else f"{port} ({tag})"
                    self.addItem(label, port)

            index = self.findData(app_model.serial_connection_port)
            self.setCurrentIndex(index)


class USBDeviceComboBox(QComboBox):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)
        self.app_model = app_model
        app_model.sig_notify.connect(self._on_app_model_update)
        self.currentTextChanged.connect(self._on_change)
        self.setMinimumWidth(175)

    def _on_change(self) -> None:
        self.app_model.set_usb_connection_port(self.currentData())

    def _on_app_model_update(self, app_model: AppModel) -> None:
        usb_devices = app_model.available_usb_devices

        view_ports = [self.itemData(i) for i in range(self.count())]

        with QtCore.QSignalBlocker(self):
            if usb_devices != view_ports:
                self.clear()
                for usb_device in usb_devices:
                    self.addItem(usb_device.name, usb_device)

            index = self.findData(app_model.usb_connection_device)
            self.setCurrentIndex(index)


class ConnectionHint(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)
        app_model.sig_notify.connect(self._on_app_model_update)

        self.setLayout(QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.icon = qta.IconWidget()
        self.icon.setIcon(qta.icon("fa.warning", color="#ff9e00"))
        self.layout().addWidget(self.icon)

        self.label = QLabel(self)
        self.label.setText("You may experience stability issues")
        self.layout().addWidget(self.label)

        self.button = QPushButton(self)
        self.button.setIcon(qta.icon("fa5s.external-link-alt", color=BUTTON_ICON_COLOR))
        self.button.setText("How to fix")
        self.button.clicked.connect(self._on_click)
        self.layout().addWidget(self.button)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        hide = not self._should_show(app_model)
        self.icon.setHidden(hide)
        self.label.setHidden(hide)
        self.button.setHidden(hide)

    def _on_click(self) -> None:
        url = r"https://docs.acconeer.com/en/latest/evk_setup/xc120_xe121.html"
        webbrowser.open_new_tab(url)

    @staticmethod
    def _should_show(app_model: AppModel) -> bool:
        if platform.system().lower() != "windows":
            return False

        if app_model.connection_interface != ConnectionInterface.SERIAL:
            return False

        port = app_model.serial_connection_port
        if port is None:
            return False

        tag = dict(app_model.available_tagged_ports).get(port, None)

        if tag is None:
            return False

        return "xc120" in tag.lower()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import copy
import logging
import os
from typing import Optional

import attrs
import numpy as np
from numpy import pi, unravel_index
from scipy.fftpack import fft, fftshift

import acconeer.exptool as et

from .calibration import ObstacleDetectionCalibration
from .constants import WAVELENGTH


BACKGROUND_PLATEAU_INTERPOLATION = 0.25  # Default plateau for background parameterization
BACKGROUND_PLATEAU_FACTOR = 1.25  # Default factor for background parameterization
MAX_SPEED = 8.00  # Max speed to be resolved with FFT in cm/s


DIR_PATH = os.path.dirname(os.path.realpath(__file__))
log = logging.getLogger(__name__)


def get_sensor_config():
    config = et.a111.IQServiceConfig()
    config.range_interval = [0.1, 0.5]
    config.repetition_mode = et.a111.IQServiceConfig.RepetitionMode.SENSOR_DRIVEN
    config.update_rate = int(np.ceil(MAX_SPEED * 4 / WAVELENGTH))
    config.gain = 0.7
    return config


class ProcessingConfiguration(et.configbase.ProcessingConfig):
    VERSION = 1

    fft_length = et.configbase.IntParameter(label="FFT length", default_value=16, limits=(2, 512))

    threshold = et.configbase.FloatParameter(
        # Ignore data below threshold in FFT window for moving objects
        label="Moving Threshold",
        default_value=0.05,
        limits=(0, 100),
        decimals=2,
        order=0,
    )

    downsampling = et.configbase.IntParameter(
        label="Downsample scale",
        default_value=1,
        limits=(0, 124),
        category=et.configbase.Category.ADVANCED,
        order=10,
    )

    calib = et.configbase.IntParameter(
        label="Background iterations",
        default_value=10,
        limits=(1, 1000),
        category=et.configbase.Category.ADVANCED,
        order=20,
    )

    bg_offset = et.configbase.FloatParameter(
        label="Background Scale",
        default_value=1.6,
        limits=(0, 1000),
        decimals=1,
        category=et.configbase.Category.ADVANCED,
        order=30,
    )

    static_threshold = et.configbase.FloatParameter(
        label="Stationary Threshold",
        default_value=0.1,
        limits=(0.0, 100),
        decimals=1,
        category=et.configbase.Category.ADVANCED,
        order=40,
    )

    close_threshold_addition = et.configbase.FloatParameter(
        # Ignore data below threshold for very close range
        label="Close Threshold Addition",
        default_value=0.1,
        limits=(0.0, 100),
        decimals=1,
        category=et.configbase.Category.ADVANCED,
        order=50,
    )

    static_distance = et.configbase.FloatParameter(
        label="Distance limit far",
        default_value=45,
        limits=(0.0, 1000),
        decimals=1,
        category=et.configbase.Category.ADVANCED,
        order=60,
    )

    static_grad = et.configbase.FloatParameter(
        label="Static distance gradient",
        default_value=6,
        limits=(0.0, 100),
        decimals=1,
        category=et.configbase.Category.ADVANCED,
        order=70,
    )

    close_dist = et.configbase.FloatParameter(
        label="Distance limit near",
        default_value=16,
        limits=(0.0, 100),
        decimals=1,
        category=et.configbase.Category.ADVANCED,
        order=80,
    )

    static_freq = et.configbase.FloatParameter(
        label="Static frequency gradient",
        default_value=2,
        limits=(0.0, 100),
        decimals=1,
        category=et.configbase.Category.ADVANCED,
        order=90,
    )

    nr_peaks = et.configbase.IntParameter(
        label="Number of peaks",
        default_value=1,
        limits=(0, 100),
        category=et.configbase.Category.ADVANCED,
        order=100,
    )

    edge_to_peak = et.configbase.FloatParameter(
        label="Edge to peak ratio",
        default_value=1,
        limits=(0.0, 1),
        decimals=1,
        category=et.configbase.Category.ADVANCED,
        order=110,
    )

    peak_hist = et.configbase.IntParameter(
        label="Peak history",
        default_value=500,
        limits=(50, 2000),
        category=et.configbase.Category.ADVANCED,
        order=120,
    )

    robot_velocity = et.configbase.FloatParameter(
        label="Robot Velocity",
        unit="cm/s",
        default_value=6,
        limits=(-1000, 1000),
        category=et.configbase.Category.ADVANCED,
        order=130,
    )

    use_parameterization = et.configbase.BoolParameter(
        label="Use bg parameterization",
        default_value=False,
        category=et.configbase.Category.ADVANCED,
        order=140,
    )

    background_map = et.configbase.BoolParameter(
        label="Show background",
        default_value=True,
        category=et.configbase.Category.ADVANCED,
        order=150,
    )

    threshold_map = et.configbase.BoolParameter(
        label="Show threshold map",
        default_value=False,
        category=et.configbase.Category.ADVANCED,
        order=160,
    )

    show_line_outs = et.configbase.BoolParameter(
        label="Show extra line outs",
        default_value=False,
        category=et.configbase.Category.ADVANCED,
        order=170,
    )

    distance_history = et.configbase.BoolParameter(
        label="Show distance history",
        default_value=True,
        category=et.configbase.Category.ADVANCED,
        order=180,
    )

    velocity_history = et.configbase.BoolParameter(
        label="Show velocity history",
        default_value=True,
        category=et.configbase.Category.ADVANCED,
        order=190,
    )

    angle_history = et.configbase.BoolParameter(
        label="Show angle history",
        default_value=True,
        category=et.configbase.Category.ADVANCED,
        order=200,
    )

    distance_history = et.configbase.BoolParameter(
        label="Show distance history",
        default_value=True,
        category=et.configbase.Category.ADVANCED,
        order=210,
    )

    amplitude_history = et.configbase.BoolParameter(
        label="Show amplitude history",
        default_value=True,
        category=et.configbase.Category.ADVANCED,
        order=220,
    )

    sensor_separation = et.configbase.FloatParameter(
        label="Sensor separation",
        unit="cm",
        default_value=15,
        limits=(1, 100),
        category=et.configbase.Category.ADVANCED,
        order=230,
    )


class Processor:
    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        self.sweep_index = 0

        self.sensor_config = sensor_config

        self.sensor_separation = processing_config.sensor_separation
        self.fft_len = processing_config.fft_length
        self.threshold = processing_config.threshold
        self.static_threshold = processing_config.static_threshold
        self.static_distance = processing_config.static_distance
        self.close_threshold_addition = processing_config.close_threshold_addition
        self.calibration_iterations_left = max(processing_config.calib, 0)
        self.calibration_iterations = self.calibration_iterations_left
        self.peak_hist_len = processing_config.peak_hist
        self.bg_off = processing_config.bg_offset
        self.nr_locals = processing_config.nr_peaks
        self.static_freq_limit = processing_config.static_freq
        self.static_dist_gradient = processing_config.static_grad
        self.close_dist_limit = processing_config.close_dist
        self.robot_velocity = processing_config.robot_velocity
        self.edge_ratio = processing_config.edge_to_peak
        self.downsampling = processing_config.downsampling

        self.len_range = session_info["data_length"]
        self._reset_calculation_arrays(self.len_range)

        if calibration is not None:
            self.update_calibration(calibration)

    def update_calibration(self, calibration: Optional[ObstacleDetectionCalibration]):
        """
        Updates the calibration of processor

        `calibration=None` is expected to remove the calibration from the processor

        Calibration will not truly change until `process` is called.
        """
        if calibration is None:
            self._reset_calibration()
            return

        calib_bg_params = attrs.asdict(copy.deepcopy(calibration))
        calib_bg_params["moving_max"] = [calib_bg_params["moving_max"]]
        calib_bg_params["static_adjacent_factor"] = [calib_bg_params["static_adjacent_factor"]]

        self._apply_calibration(calib_bg_params)

    def _reset_calibration(self):
        """
        Will cause `self` to enter `_first_sweep_setup` on the next `process`-call,
        which will reset most of the state.
        A new calibration will then begin since `self.bg_avg` and `self.sweep_index` is set to 0.
        """
        self.sweep_index = 0
        self.calibration_iterations_left = self.calibration_iterations
        self.is_calibrated = False
        self._reset_calculation_arrays(self.len_range)

    def _apply_calibration(self, parameterized_calibration):
        """
        Will populate self.fft_bg with a "raw" background that is calculated
        from the passed parameterization.
        """
        if not isinstance(parameterized_calibration, dict):
            log.warning("Received unsupported background data!")
            return

        try:
            # Generate background from piece-wise-linear (pwl) interpolation
            self.generate_background_from_pwl_params(
                self.fft_bg[0, :, :], parameterized_calibration
            )
            self.calibration_iterations_left = 0
            self.is_calibrated = True
            log.info("Using saved parameterized FFT background data!")
        except Exception:
            log.warning("Could not reconstruct background!")

    def _reset_calculation_arrays(self, len_range):
        """Resets all arrays used for calculations. This was done on first sweep before."""
        self.sweep_map = np.zeros((1, len_range, self.fft_len), dtype="complex")
        self.fft_bg = np.zeros((1, len_range, self.fft_len))
        self.hamming_map = np.zeros((len_range, self.fft_len))

        for i in range(len_range):
            self.hamming_map[i, :] = np.hamming(self.fft_len)

        self.env_xs = np.linspace(*self.sensor_config.range_interval * 100, len_range)
        self.peak_prop_num = 4
        self.peak_hist = np.zeros((1, self.nr_locals, self.peak_prop_num, self.peak_hist_len))
        self.peak_hist *= float(np.nan)
        self.mask = np.zeros((len_range, self.fft_len))
        self.threshold_map = np.zeros((len_range, self.fft_len))

        for dist in range(len_range):
            for freq in range(self.fft_len):
                self.threshold_map[dist, freq] = self.variable_thresholding(
                    freq, dist, self.threshold, self.static_threshold
                )

    def _process_single_sensor(self, sweep, fft_psd):
        self.push(sweep[0, :], self.sweep_map[0, :, :])

        signalFFT = fftshift(fft(self.sweep_map[0, :, :] * self.hamming_map, axis=1), axes=1)
        fft_psd[0, :, :] = np.square(np.abs(signalFFT))
        signalPSD = fft_psd[0, :, :]

        out_data = {}

        if self.calibration_iterations_left > 0 and self.sweep_index == self.fft_len - 1:
            self.fft_bg[0, :, :] = np.maximum(self.bg_off * signalPSD, self.fft_bg[0, :, :])
            self.calibration_iterations_left -= 1
            if self.calibration_iterations_left == 0:
                bg_params = self.parameterize_bg(self.fft_bg[0, :, :])
                out_data["new_calibration"] = ObstacleDetectionCalibration(**bg_params)

        signalPSD_sub = signalPSD - self.fft_bg[0, :, :]
        signalPSD_sub[signalPSD_sub < 0] = 0
        env = np.abs(sweep[0, :])

        fft_peaks, peaks_found = self.find_peaks(signalPSD_sub)
        fft_max_env = signalPSD[:, 8]
        angle = None
        velocity = None
        peak_idx = np.argmax(env)

        if self.sweep_index < self.fft_len:
            fft_peaks = None

        if fft_peaks is not None:
            fft_max_env = signalPSD[:, int(fft_peaks[0, 1])]
            zero = np.floor(self.fft_len / 2)

            for i in range(self.nr_locals):
                bin_index = fft_peaks[i, 2] - zero
                velocity = (bin_index / zero) * WAVELENGTH * self.sensor_config.update_rate / 4
                angle = np.arccos(self.clamp(abs(velocity) / self.robot_velocity, -1.0, 1.0))
                angle = np.sign(velocity) * angle / pi * 180
                peak_idx = int(fft_peaks[i, 0])
                distance = self.env_xs[int(fft_peaks[i, 0])]
                amp = fft_peaks[i, 3]

                if not amp:
                    distance = float(np.nan)
                    velocity = float(np.nan)
                    angle = float(np.nan)
                    amp = float(np.nan)

                self.push_vec(distance, self.peak_hist[0, i, 0, :])
                self.push_vec(velocity, self.peak_hist[0, i, 1, :])
                self.push_vec(angle, self.peak_hist[0, i, 2, :])
                self.push_vec(amp, self.peak_hist[0, i, 3, :])

            fft_peaks = fft_peaks[:peaks_found, :]
        else:
            for i in range(self.nr_locals):
                for j in range(self.peak_prop_num):
                    self.push_vec(float(np.nan), self.peak_hist[0, i, j, :])

        out_data.update(
            {
                "env_ampl": env,
                "fft_max_env": fft_max_env,
                "fft_map": fft_psd,
                "peak_idx": peak_idx,
                "angle": angle,
                "velocity": velocity,
                "fft_peaks": fft_peaks,
                "peak_hist": self.peak_hist[0, :, :, :],
                "sweep_index": self.sweep_index,
                "peaks_found": peaks_found,
            }
        )
        return out_data

    def process(self, data, data_info):
        sweep = data

        if len(sweep.shape) == 1:
            sweep = np.expand_dims(sweep, 0)

        if self.downsampling:
            sweep = sweep[:, :: self.downsampling]

        sweep = sweep / 2**12

        _, len_range = sweep.shape

        fft_psd = np.empty((1, len_range, self.fft_len))

        out_data = self._process_single_sensor(sweep, fft_psd)

        fft_bg = None
        fft_bg_send = None
        if self.sweep_index < 3 * self.fft_len:  # Make sure data gets send to ui
            fft_bg = self.fft_bg[0, :, :]
            fft_bg_send = self.fft_bg

        threshold_map = None
        if self.sweep_index < 3 * self.fft_len:  # Make sure data gets send to ui
            threshold_map = self.threshold_map

        out_data.update(
            {
                "fft_bg": fft_bg,
                "fft_bg_iterations_left": self.calibration_iterations_left,
                "threshold_map": threshold_map,
                "send_process_data": fft_bg_send,
            }
        )

        if self.calibration_iterations_left > 0 and self.sweep_index == self.fft_len - 1:
            self.sweep_index = 0
        else:
            self.sweep_index += 1

        return out_data

    def remap(self, val, x1, x2, y1, y2):
        if x1 == x2:
            return y1
        m = (y2 - y1) / (x2 - x1)
        b = y1 - m * x1
        return val * m + b

    def clamp(self, val, a, b):
        val = max(val, a)
        val = min(val, b)
        return val

    def push(self, sweep, arr):
        res = np.empty_like(arr)
        res[:, 0] = sweep
        res[:, 1:] = arr[:, :-1]
        arr[...] = res

    def push_vec(self, val, vec):
        res = np.empty_like(vec)
        res[0] = val
        res[1:] = vec[:-1]
        vec[...] = res

    def parameterize_bg(self, fft_bg):
        dist_len = fft_bg.shape[0]

        static_idx = int(self.fft_len / 2)
        adjacent_idx = [int(static_idx - 1), int(static_idx + 1)]
        moving_range = [*range(adjacent_idx[0]), *range(adjacent_idx[1] + 1, self.fft_len)]
        pwl_static_points = 5

        static_pwl_amp = []
        static_pwl_dist = []

        static_sum = 0.0
        adjacent_sum = np.zeros(2)
        moving_sum = 0.0
        moving_sum_sqr = 0.0
        moving_mean_array = np.zeros(dist_len)

        moving_max = 0.0
        moving_min = 0.0

        for dist_index in range(dist_len):
            static_sum += fft_bg[dist_index, static_idx]
            adjacent_sum[0] += fft_bg[dist_index, adjacent_idx[0]]
            adjacent_sum[1] += fft_bg[dist_index, adjacent_idx[1]]

            current_dist_moving_sum = 0.0
            for freq_index in moving_range:
                val = fft_bg[dist_index, freq_index]
                current_dist_moving_sum += val
                if val > moving_max:
                    moving_max = val
                if val < moving_min:
                    moving_min = val
                moving_sum += val
                moving_sum_sqr += val * val

            moving_mean_array[dist_index] = current_dist_moving_sum / len(moving_range)

        segs = self.adapt_background_segment_step(moving_mean_array, dist_len)
        segs_nr = int(len(segs) / 2)
        moving_pwl_dist = segs[0:segs_nr]
        moving_pwl_amp = segs[segs_nr:]

        adjacent_sum_max = max(adjacent_sum)

        if adjacent_sum_max == 0:
            static_adjacent_factor = 0
        else:
            static_adjacent_factor = static_sum / adjacent_sum_max

        bin_width = dist_len / pwl_static_points

        pwl_y_max = 0
        pwl_x_max = 0
        for point_index in range(pwl_static_points):
            dist_begin = int(point_index * bin_width)
            dist_end = int((point_index + 1) * bin_width)

            if point_index == pwl_static_points:
                dist_end = dist_len

            pwl_x_max = 0.0
            pwl_y_max = 0.0
            for dist_index in range(dist_begin, dist_end):
                val = fft_bg[dist_index, static_idx]
                if val > pwl_y_max:
                    pwl_x_max = dist_index
                    pwl_y_max = val

            static_pwl_dist.append(pwl_x_max)
            static_pwl_amp.append(pwl_y_max)

        for i in range(pwl_static_points):
            if static_pwl_amp[i] < moving_max:
                static_pwl_amp[i] = moving_max

        bg_params = {
            "static_pwl_dist": [float(i) for i in static_pwl_dist],
            "static_pwl_amp": [float(i) for i in static_pwl_amp],
            "moving_pwl_dist": [float(i) for i in moving_pwl_dist],
            "moving_pwl_amp": [float(i) for i in moving_pwl_amp],
            "static_adjacent_factor": [float(static_adjacent_factor)],
            "moving_max": [float(moving_max)],
        }

        return bg_params

    def adapt_background_segment_step(self, data_y, data_length):
        mid_index = int(data_length / 2)

        y1 = 0
        for i in range(mid_index):
            if y1 < data_y[i]:
                y1 = data_y[i]

        y2 = 0
        for i in range(mid_index, int(data_length)):
            if y2 < data_y[i]:
                y2 = data_y[i]

        # Check the plateau levels and return early if the step is non-decreasing
        if y1 <= y2:
            y1 = y2
            x1 = 0
            x2 = float(data_length - 1)
            return [x1, x2, y1, y2]

        intersection_threshold = self.remap(BACKGROUND_PLATEAU_INTERPOLATION, 0, 1, y2, y1)
        if intersection_threshold > y2 * BACKGROUND_PLATEAU_FACTOR:
            intersection_threshold = y2 * BACKGROUND_PLATEAU_FACTOR

        intersection_index = mid_index
        while intersection_index > 1 and data_y[intersection_index - 1] < intersection_threshold:
            intersection_index -= 1

        x2 = float(intersection_index)
        y2 = data_y[intersection_index]

        if intersection_index <= 1:
            x1 = 0
            return [x1, x2, y1, y2]

        max_neg_slope = 0
        for i in range(intersection_index):
            neg_slope = (data_y[i] - y2) / float(intersection_index - i)
            if max_neg_slope < neg_slope:
                max_neg_slope = neg_slope
        x1 = x2 - (y1 - y2) / max_neg_slope

        return [x1, x2, y1, y2]

    def generate_background_from_pwl_params(self, fft_bg, bg_params):
        static_idx = int(self.fft_len / 2)
        adjacent_idx = [int(static_idx - 1), int(static_idx + 1)]
        moving_range = [*range(adjacent_idx[0]), *range(adjacent_idx[1] + 1, self.fft_len)]
        pwl_static_points = 5
        pwl_moving_points = 2
        dist_len = fft_bg.shape[0]
        fac = bg_params["static_adjacent_factor"][0]
        static_pwl_amp = bg_params["static_pwl_amp"]
        static_pwl_dist = bg_params["static_pwl_dist"]
        moving_pwl_amp = bg_params["moving_pwl_amp"]
        moving_pwl_dist = bg_params["moving_pwl_dist"]
        moving_max = bg_params["moving_max"][0]

        self.apply_pwl_segments(
            fft_bg, static_idx, static_pwl_dist, static_pwl_amp, pwl_static_points
        )

        for dist_index in range(dist_len):
            static_val = fft_bg[dist_index, 8]
            if static_val < moving_max:
                static_val = moving_max
                fft_bg[dist_index, static_idx] = static_val

            interp_adjacent = moving_max
            if fac > 0:
                interp_adjacent = static_val / fac
                fft_bg[dist_index, adjacent_idx] = interp_adjacent

        for freq_index in moving_range:
            self.apply_pwl_segments(
                fft_bg, freq_index, moving_pwl_dist, moving_pwl_amp, pwl_moving_points
            )

    def apply_pwl_segments(self, fft_bg, freq_index, pwl_dist, pwl_amp, pwl_points):
        dist_len = fft_bg.shape[0]

        x_start = 0
        x_stop = pwl_dist[0]
        y_start = pwl_amp[0]
        y_stop = pwl_amp[0]

        segment_stop = 0
        for dist_index in range(dist_len):
            if x_stop < dist_index:
                x_start = x_stop
                y_start = y_stop
                segment_stop += 1
                if segment_stop < pwl_points:
                    x_stop = pwl_dist[segment_stop]
                    y_stop = pwl_amp[segment_stop]
                else:
                    x_stop = dist_len - 1
            interp = self.remap(dist_index, x_start, x_stop, y_start, y_stop)
            fft_bg[dist_index, freq_index] = interp

    def find_peaks(self, arr):
        if not self.nr_locals:
            return None, 0

        peaks_found = 0
        peak = np.asarray(unravel_index(np.argmax(arr), arr.shape))
        peak_avg = peak[1]

        peak_val = arr[peak[0], peak[1]]
        thresh = self.variable_thresholding(
            peak[1], peak[0], self.threshold, self.static_threshold
        )

        if peak_val < thresh:
            peak = None
            peak_avg = None

        if peak is not None:
            peak_avg = self.calc_peak_avg(peak, arr)
            local_peaks = np.zeros((self.nr_locals, self.peak_prop_num), dtype=float)
            dist_edge = self.edge(arr[:, peak[1]], peak[0], self.edge_ratio)
            local_peaks[0, 0] = dist_edge
            local_peaks[0, 1] = peak[1]
            local_peaks[0, 2] = peak_avg
            local_peaks[0, 3] = np.round(peak_val, decimals=4)
            peaks_found += 1
            self.mask = arr.copy()

            for i in range(self.nr_locals - 1):
                self.peak_masking(local_peaks[i, :])
                p = np.asarray(unravel_index(np.argmax(self.mask), arr.shape))
                thresh = self.variable_thresholding(
                    p[1], p[0], self.threshold, self.static_threshold
                )
                peak_val = arr[p[0], p[1]]
                if peak_val > thresh:
                    dist_edge = self.edge(arr[:, p[1]], p[0], self.edge_ratio)
                    local_peaks[i + 1, 0] = dist_edge
                    local_peaks[i + 1, 1] = p[1]
                    local_peaks[i + 1, 2] = self.calc_peak_avg(p, arr)
                    local_peaks[i + 1, 3] = np.round(peak_val, decimals=4)
                    peaks_found += 1
                else:
                    break
            return local_peaks, peaks_found
        else:
            return None, 0

    def calc_peak_avg(self, peak, arr):
        amp_sum = 0
        s = 0
        for i in range(3):
            idx_real = peak[1] - 1 + i
            idx = idx_real
            if idx == self.fft_len:
                idx = 0
            elif idx == -1:
                idx = self.fft_len - 1
            s += arr[peak[0], idx] * idx_real
            amp_sum += arr[peak[0], idx]
        peak_avg = s / amp_sum

        peak_avg = max(0, peak_avg)

        return peak_avg % self.fft_len

    def peak_masking(self, peak, angle_depth=2, depth=180, amplitude_margin=1.2):
        dist_depth = depth / self.downsampling

        distance_index = peak[0]
        angle_index = peak[1]
        peak_val = peak[3]

        dist_len, angle_len = self.mask.shape

        distance_scaling_per_index = 1 / dist_depth
        angle_scaling_per_index = 1 / (1 + angle_depth)

        distance_start_index = -dist_depth

        if distance_start_index + distance_index < 0:
            distance_start_index = -distance_index

        distance_end_index = dist_depth
        if distance_end_index + distance_index >= dist_len:
            distance_end_index = dist_len - distance_index - 1

        for i in range(-angle_depth, angle_depth + 1):
            wrapped_row = (angle_len + i + angle_index) % angle_len
            for j in range(int(distance_start_index), int(distance_end_index) + 1):
                dist_from_peak = (
                    abs(i) * angle_scaling_per_index + abs(j) * distance_scaling_per_index
                )
                mask_val = (1 - dist_from_peak**2) * peak_val * amplitude_margin
                if self.mask[int(j + distance_index), int(wrapped_row)] < mask_val:
                    self.mask[int(j + distance_index), int(wrapped_row)] = 0

    def edge(self, arr, peak_idx, ratio=0.5):
        if ratio == 1.0:
            return peak_idx

        s0 = arr[peak_idx]
        for i in range(peak_idx):
            if peak_idx - i < 0:
                peak_idx = 0
                break
            if arr[peak_idx - i] < s0 * ratio:
                peak_idx -= i
                break

        return peak_idx

    def variable_thresholding(self, freq_index, dist_index, min_thresh, max_thresh):
        dist = self.env_xs[dist_index]
        distance_gradient = self.static_dist_gradient
        thresh = self.remap(
            dist,
            self.static_distance - distance_gradient,
            self.static_distance,
            max_thresh,
            min_thresh,
        )
        thresh = self.clamp(thresh, min_thresh, max_thresh)

        null_frequency = self.fft_len / 2
        frequency_gradient = self.static_freq_limit

        if freq_index <= null_frequency:
            freq = self.clamp(freq_index, null_frequency - frequency_gradient, null_frequency)
            thresh = self.remap(
                freq, null_frequency - frequency_gradient, null_frequency, min_thresh, thresh
            )
        else:
            freq = self.clamp(freq_index, null_frequency, null_frequency + frequency_gradient)
            thresh = self.remap(
                freq, null_frequency, null_frequency + frequency_gradient, thresh, min_thresh
            )

        thresh_add = self.remap(
            dist,
            self.sensor_config.range_start * 100,
            self.sensor_config.range_start * 100 + self.close_dist_limit,
            self.close_threshold_addition,
            0.0,
        )

        thresh += self.clamp(thresh_add, 0, self.close_threshold_addition)

        return thresh



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import json
from typing import Any, Optional, Tuple, Union

import attrs
import numpy as np
from typing_extensions import Literal, NotRequired, TypedDict

from acconeer.exptool.a121._core.entities import (
    INT_16_COMPLEX,
    PRF,
    IdleState,
    Metadata,
    Result,
    ResultContext,
    SensorInfo,
    ServerInfo,
    SessionConfig,
)
from acconeer.exptool.a121._core.mediators import CommunicationProtocol
from acconeer.exptool.a121._core.utils import map_over_extended_structure


GoodStatus = Union[Literal["ok"], Literal["start"], Literal["stop"], Literal["end"]]
BadStatus = Literal["error"]


class ValidResponse(TypedDict):
    status: GoodStatus


class ErroneousResponse(TypedDict):
    status: BadStatus
    message: str


AnyResponse = Union[ValidResponse, ErroneousResponse]


class SystemInfo(TypedDict):
    rss_version: str
    sensor: str
    sensor_count: int
    ticks_per_second: int
    hw: NotRequired[Optional[str]]


class GetSystemInfoResponse(ValidResponse):
    system_info: SystemInfo


class SensorInfoDict(TypedDict):
    connected: bool
    serial: Optional[str]


class GetSensorInfoResponse(ValidResponse):
    sensor_info: list[SensorInfoDict]


class MetadataResponse(TypedDict):
    frame_data_length: int
    sweep_data_length: int
    subsweep_data_offset: list[int]
    subsweep_data_length: list[int]
    calibration_temperature: int
    base_step_length_m: float
    max_sweep_rate: float


class SetupResponse(ValidResponse):
    tick_period: int
    metadata: list[list[MetadataResponse]]


class ResultInfoDict(TypedDict):
    tick: int
    data_saturated: bool
    frame_delayed: bool
    calibration_needed: bool
    temperature: int


class GetNextHeader(ValidResponse):
    result_info: list[list[ResultInfoDict]]
    payload_size: int


class ExplorationProtocolError(Exception):
    pass


class ServerError(Exception):
    pass


class ExplorationProtocol(CommunicationProtocol):
    end_sequence: bytes = b"\n"

    PRF_MAPPING = {
        PRF.PRF_19_5_MHz: "19_5_MHz",
        PRF.PRF_13_0_MHz: "13_0_MHz",
        PRF.PRF_8_7_MHz: "8_7_MHz",
        PRF.PRF_6_5_MHz: "6_5_MHz",
    }
    IDLE_STATE_MAPPING = {
        IdleState.DEEP_SLEEP: "deep_sleep",
        IdleState.SLEEP: "sleep",
        IdleState.READY: "ready",
    }

    @staticmethod
    def check_status(response: AnyResponse, expected: GoodStatus) -> None:
        if response["status"] == "error":
            raise ServerError(response["message"])

        if response["status"] != expected:
            raise ServerError(
                f'Unexpected reponse status from server. Was "{response}", '
                + f'expected "{expected}"'
            )

    @classmethod
    def get_system_info_command(cls) -> bytes:
        return b'{"cmd":"get_system_info"}\n'

    @classmethod
    def get_system_info_response(
        cls, bytes_: bytes, sensor_infos: dict[int, SensorInfo]
    ) -> Tuple[ServerInfo, str]:
        response: GetSystemInfoResponse = json.loads(bytes_)
        cls.check_status(response, expected="ok")

        try:
            system_info = response["system_info"]
            return (
                ServerInfo(
                    rss_version=system_info["rss_version"],
                    sensor_count=system_info["sensor_count"],
                    ticks_per_second=system_info["ticks_per_second"],
                    sensor_infos=sensor_infos,
                    hardware_name=system_info.get("hw"),
                ),
                system_info["sensor"],
            )
        except KeyError as ke:
            raise ExplorationProtocolError(
                f"Could not parse 'get_system_info' response. Response: {response}"
            ) from ke

    @classmethod
    def get_sensor_info_command(cls) -> bytes:
        return b'{"cmd":"get_sensor_info"}\n'

    @classmethod
    def get_sensor_info_response(cls, bytes_: bytes) -> dict[int, SensorInfo]:
        response: GetSensorInfoResponse = json.loads(bytes_)
        cls.check_status(response, expected="ok")

        sensor_infos = response["sensor_info"]

        return {
            sensor_id: SensorInfo(
                connected=sensor_info_dict["connected"], serial=sensor_info_dict.get("serial")
            )
            for sensor_id, sensor_info_dict in enumerate(sensor_infos, start=1)
        }

    @classmethod
    def _setup_command_preprocessing(cls, session_config: SessionConfig) -> dict:
        result = session_config.to_dict()

        if session_config.update_rate is None:
            del result["update_rate"]

        # Exploration server is not interested in this.
        result.pop("extended")

        result["groups"] = map_over_extended_structure(cls._translate_prf_enums, result["groups"])
        result["groups"] = map_over_extended_structure(
            cls._translate_profile_enums, result["groups"]
        )
        result["groups"] = map_over_extended_structure(
            cls._translate_idle_state_enums, result["groups"]
        )
        result["groups"] = map_over_extended_structure(
            cls._translate_sentiel_values, result["groups"]
        )

        result["cmd"] = "setup"
        return result

    @classmethod
    def setup_command(cls, session_config: SessionConfig) -> bytes:
        result = cls._setup_command_preprocessing(session_config)
        result["groups"] = cls._translate_groups_representation(result["groups"])
        return (
            json.dumps(
                result,
                separators=(",", ":"),
                ensure_ascii=True,
            )
            + "\n"
        ).encode("ascii")

    @classmethod
    def _translate_profile_enums(cls, sensor_config_dict: dict[str, Any]) -> dict[str, Any]:
        for subsweep_config_dict in sensor_config_dict["subsweeps"]:
            subsweep_config_dict["profile"] = subsweep_config_dict["profile"].value
        return sensor_config_dict

    @classmethod
    def _translate_prf_enums(cls, sensor_config_dict: dict[str, Any]) -> dict[str, Any]:
        for subsweep_config_dict in sensor_config_dict["subsweeps"]:
            subsweep_config_dict["prf"] = cls.PRF_MAPPING[subsweep_config_dict["prf"]]
        return sensor_config_dict

    @classmethod
    def _translate_idle_state_enums(cls, sensor_config_dict: dict[str, Any]) -> dict[str, Any]:
        sensor_config_dict["inter_frame_idle_state"] = cls.IDLE_STATE_MAPPING[
            sensor_config_dict["inter_frame_idle_state"]
        ]
        sensor_config_dict["inter_sweep_idle_state"] = cls.IDLE_STATE_MAPPING[
            sensor_config_dict["inter_sweep_idle_state"]
        ]
        return sensor_config_dict

    @classmethod
    def _translate_sentiel_values(cls, sensor_config_dict: dict[str, Any]) -> dict[str, Any]:
        """'sweep_rate' and 'frame_rate' has unlocked rate when they are = None.
        RSS's sentinel is 0.0 (zero)
        """
        sensor_config_dict["sweep_rate"] = sensor_config_dict["sweep_rate"] or 0.0
        sensor_config_dict["frame_rate"] = sensor_config_dict["frame_rate"] or 0.0
        return sensor_config_dict

    @classmethod
    def _translate_groups_representation(
        cls, groups_list: list[dict[int, Any]]
    ) -> list[list[dict[str, Any]]]:
        """
        This function translates the Exptool representation, which is

        groups = [
            {sensor_id1: config1, sensor_id2: config2, ...},  # Group 1
            ...,
        ]

        To the representation the Exploration server expects;

        groups = [
            [  # Group 1
                {"sensor_id": sensor_id1, "config": config1},
                {"sensor_id": sensor_id2, "config": config2},
                ...
            ],
            ...
        ]

        """
        return [
            [{"sensor_id": sensor_id, "config": config} for sensor_id, config in group.items()]
            for group in groups_list
        ]

    @classmethod
    def setup_response(
        cls, bytes_: bytes, context_session_config: SessionConfig
    ) -> list[dict[int, Metadata]]:
        response: SetupResponse = json.loads(bytes_)
        cls.check_status(response, expected="ok")

        metadata_groups = response["metadata"]

        result = []

        for metadata_group, config_group in zip(metadata_groups, context_session_config._groups):
            # OBS! This logic relies on insertion order.
            # Bugs may arise where meta-data is unaligned with configs.
            result.append(
                {
                    sensor_id: cls._create_metadata(metadata_dict, response["tick_period"])
                    for metadata_dict, sensor_id in zip(metadata_group, config_group.keys())
                }
            )

        return result

    @classmethod
    def _create_metadata(cls, metadata_dict: MetadataResponse, tick_period: int) -> Metadata:
        return Metadata(
            frame_data_length=metadata_dict["frame_data_length"],
            sweep_data_length=metadata_dict["sweep_data_length"],
            subsweep_data_offset=np.array(metadata_dict["subsweep_data_offset"]),
            subsweep_data_length=np.array(metadata_dict["subsweep_data_length"]),
            calibration_temperature=metadata_dict["calibration_temperature"],
            tick_period=tick_period,
            base_step_length_m=metadata_dict["base_step_length_m"],
            max_sweep_rate=metadata_dict["max_sweep_rate"],
        )

    @classmethod
    def start_streaming_command(cls) -> bytes:
        return b'{"cmd":"start_streaming"}\n'

    @classmethod
    def start_streaming_response(cls, bytes_: bytes) -> None:
        response: AnyResponse = json.loads(bytes_)
        cls.check_status(response, "start")

    @classmethod
    def stop_streaming_command(cls):
        return b'{"cmd":"stop_streaming"}\n'

    @classmethod
    def stop_streaming_response(cls, bytes_: bytes) -> None:
        response: AnyResponse = json.loads(bytes_)
        cls.check_status(response, "stop")

    @classmethod
    def get_next_header(
        cls, bytes_: bytes, extended_metadata: list[dict[int, Metadata]], ticks_per_second: int
    ) -> Tuple[int, list[dict[int, Result]]]:
        header_dict: GetNextHeader = json.loads(bytes_)
        cls.check_status(header_dict, expected="ok")

        payload_size = header_dict["payload_size"]

        extended_partial_results = []
        partial_result_groups = header_dict["result_info"]

        for partial_result_group, metadata_group in zip(partial_result_groups, extended_metadata):
            inner_result = {}
            for partial_result_dict, (sensor_id, metadata) in zip(
                partial_result_group, metadata_group.items()
            ):
                inner_result[sensor_id] = cls._create_partial_result(
                    **partial_result_dict, metadata=metadata, ticks_per_second=ticks_per_second
                )

            extended_partial_results.append(inner_result)

        return payload_size, extended_partial_results

    @classmethod
    def _create_partial_result(
        cls,
        *,
        tick: int,
        ticks_per_second: int,
        data_saturated: bool,
        temperature: int,
        frame_delayed: bool,
        calibration_needed: bool,
        metadata: Metadata,
    ) -> Result:
        return Result(
            tick=tick,
            data_saturated=data_saturated,
            frame=np.array([0]),
            temperature=temperature,
            frame_delayed=frame_delayed,
            calibration_needed=calibration_needed,
            context=ResultContext(
                metadata=metadata,
                ticks_per_second=ticks_per_second,
            ),
        )

    @classmethod
    def get_next_payload(
        cls, bytes_: bytes, partial_results: list[dict[int, Result]]
    ) -> list[dict[int, Result]]:
        start = 0
        for partial_group in partial_results:
            for sensor_id, partial_result in partial_group.items():
                metadata = partial_result._context.metadata
                end = start + metadata.frame_data_length * 4  # 4 = sizeof(int_16_complex)

                raw_frame_data = bytes_[start:end]
                assert len(raw_frame_data) == end - start
                np_frame = np.frombuffer(raw_frame_data, dtype=INT_16_COMPLEX)
                resized_frame = np.resize(np_frame, metadata.frame_shape)
                partial_group[sensor_id] = attrs.evolve(partial_result, frame=resized_frame)
                start = end
        return partial_results



from .winusberror import WinUSBError
from .winusbutils import *


class WinUSBApi(object):
    """Facade class wrapping USB library WinUSB"""

    def __init__(self):

        try:
            self._kernel32 = windll.kernel32
        except WindowsError:
            raise WinUSBError("Kernel32 dll is not present. Are you really using Windows?")

        try:
            self._winusb = windll.winusb
        except WindowsError:
            raise WinUSBError("WinUsb dll is not present")

        try:
            self._setupapi = windll.SetupApi
        except WindowsError:
            raise WinUSBError("SetupApi dll is not present")

        self._winusb_functions_dict = get_winusb_functions(self._winusb)
        self._kernel32_functions_dict = get_kernel32_functions(self._kernel32)
        self._setupapi_functions_dict = get_setupapi_functions(self._setupapi)

    def exec_function_winusb(self, function_name, *args):
        function_caller = self._configure_ctype_function(
            self._winusb_functions_dict, function_name
        )
        return function_caller(args)

    def exec_function_kernel32(self, function_name, *args):
        function_caller = self._configure_ctype_function(
            self._kernel32_functions_dict, function_name
        )
        return function_caller(args)

    def exec_function_setupapi(self, function_name, *args):
        function_caller = self._configure_ctype_function(
            self._setupapi_functions_dict, function_name
        )
        return function_caller(args)

    def _configure_ctype_function(self, dll_dict_functions, function_name):
        def _function_caller(*args):
            function = dll_dict_functions["functions"][function_name]
            try:
                function.restype = dll_dict_functions["restypes"][function_name]
                function.argtypes = dll_dict_functions["argtypes"][function_name]
            except KeyError:
                pass
            return function(*args[0])

        return _function_caller



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from acconeer.exptool.setup.base import PlatformInstall, SetupGroup, ShellCommandStep, utils

from .linux import Linux


@PlatformInstall.register
class Ubuntu_20_04(Linux):
    def __init__(self) -> None:
        super().__init__()
        self.add_steps(
            utils.WithDescription(
                '> Download a dependency via "apt" that is required '
                + "to run the Exploration tool app.",
                SetupGroup(
                    ShellCommandStep("sudo apt update".split()),
                    ShellCommandStep("sudo apt install -y libxcb-xinerama0-dev".split()),
                ),
            ),
        )

    @classmethod
    def get_key(cls) -> str:
        return "Ubuntu_20.04"



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np

import acconeer.exptool as et

from .constants import HISTORY_LENGTH_S


A_WEEK_MINUTES = 10080.0
SENSITIVITY_MAX = 1.0
MAX_ALLOWED_NOISE = 20
LP_AVERAGE_CONST_MAX = 0.9
LP_AVERAGE_CONST_INIT = 0.7

DOUBLE_PRESS_RATIO = 1.3  # Ratio between lp_constants for the trigger and cool down filters
THRESHOLD_RATIO = 2.0  # Ratio between trigger and cool down thresholds.

# Constants for calculating the number of intervals actually returned
SPARSE_RESOLUTION = 0.06
EPSILON = 0.00001


def get_sensor_config():
    sensor_config = et.a111.SparseServiceConfig()
    sensor_config.range_interval = [0.06, 0.3]
    sensor_config.sweeps_per_frame = 32
    sensor_config.update_rate = 80
    sensor_config.hw_accelerated_average_samples = 32
    sensor_config.sampling_mode = sensor_config.SamplingMode.A
    sensor_config.profile = sensor_config.Profile.PROFILE_2
    sensor_config.gain = 0.0
    sensor_config.maximize_signal_attenuation = False
    sensor_config.downsampling_factor = 1
    sensor_config.tx_disable = False
    sensor_config.power_save_mode = sensor_config.PowerSaveMode.SLEEP
    sensor_config.repetition_mode = sensor_config.RepetitionMode.HOST_DRIVEN
    return sensor_config


class ProcessingConfiguration(et.configbase.ProcessingConfig):

    recalibration_period = et.configbase.FloatParameter(
        label="Time between recalibrations",
        default_value=10000.0,
        limits=(0.1, A_WEEK_MINUTES),
        unit="min",
        logscale=True,
        updateable=True,
        order=10,
        help="How often the algorithm recalibrates itself",
    )

    sensitivity = et.configbase.FloatParameter(
        label="Detection sensitivity",
        default_value=0.7,
        limits=(0.0, SENSITIVITY_MAX),
        logscale=False,
        updateable=True,
        order=20,
        help="Sensitivity for how easily we detect a button press. "
        "Increasing the value increases the risk of false detects.",
    )

    double_press_speed = et.configbase.FloatParameter(
        label="Double press sensitivity",
        default_value=0.9,
        limits=(0.4, 0.99),
        logscale=False,
        updateable=True,
        order=30,
        help="Sensitivity for double presses. Lower value means more likely to double press."
        "High value in combination with low sensitivity can also cause 'double press' behaviour.",
    )

    def sparse_round_down(self, num):
        """
        Downwards rounding to nearest multiple of 0.06 (sparse resolution in cm)
        """
        return SPARSE_RESOLUTION * round((num - EPSILON) / SPARSE_RESOLUTION)

    def sparse_round_up(self, num):
        """
        Upwards rounding to nearest multiple of 0.06 (sparse resolution in cm)
        """
        return SPARSE_RESOLUTION * round((num + EPSILON) / SPARSE_RESOLUTION)

    def check_sensor_config(self, sensor_config):

        alerts = {
            "processing": [],
            "sensor": [],
        }

        if self.sparse_round_down(sensor_config.range_interval[0]) == 0.0:
            alerts["sensor"].append(
                et.configbase.Warning(
                    "range_interval",
                    "0.0 m point in range, increase range start for better performance",
                )
            )

        if self.recalibration_period == A_WEEK_MINUTES:
            alerts["processing"].append(
                et.configbase.Info("recalibration_period", "Recalibration Off")
            )

        return alerts


class Processor:
    def __init__(self, sensor_config, processing_config, session_info, calibration=None):

        pc = ProcessingConfiguration()
        (rounded_bottom, rounded_top) = (
            pc.sparse_round_down(sensor_config.range_interval[0]),
            pc.sparse_round_up(sensor_config.range_interval[1]),
        )
        num_depths = int(round((rounded_top - rounded_bottom) / SPARSE_RESOLUTION)) + 1

        self.num_depths = num_depths
        self.f = sensor_config.update_rate
        self.signal_history = np.zeros((int(round(self.f * HISTORY_LENGTH_S)), num_depths + 1))
        self.average_history = np.zeros((int(round(self.f * HISTORY_LENGTH_S)), num_depths + 1))
        self.trig_history = np.zeros((int(round(self.f * HISTORY_LENGTH_S)), num_depths + 1))
        self.cool_down_history = np.zeros((int(round(self.f * HISTORY_LENGTH_S)), num_depths + 1))

        self.detection_history = []
        self.frame_count = 0
        self.sweeps_per_frame = sensor_config.sweeps_per_frame

        self.calibration_limit = max(
            1, int(round(sensor_config.update_rate))
        )  # So we always wait one second. This could be more sophisticated.
        self.noise_level_target = MAX_ALLOWED_NOISE

        self.calibrated = 0  # Repeated sending of calibrated signal to exptool plot thread

        self.chilled = [True] * num_depths
        self.initialized = [False] * num_depths

        # Low pass trig
        self.lp_trig = [0.0] * num_depths
        self.lp_trig_const = [0.0] * num_depths

        # Low pass cool down
        self.lp_cool_down = [0.0] * num_depths
        self.lp_cool_down_const = [0.0] * num_depths

        # Low pass average
        self.lp_average_const = [LP_AVERAGE_CONST_INIT] * num_depths

        self.lp_average = [0.0] * num_depths

        self.update_processing_config(processing_config)

    def update_processing_config(self, processing_config):
        self.sensitivity = processing_config.sensitivity
        self.double_press_speed = processing_config.double_press_speed
        if processing_config.recalibration_period == A_WEEK_MINUTES:
            self.recalibration_frame_period = -1  # Effectively turning it off
        else:
            self.recalibration_frame_period = int(
                round(processing_config.recalibration_period * 60 * self.f)
            )

        self.set_sensitivity()
        self.set_double_press_speed()

    def set_double_press_speed(self):
        for i in range(self.num_depths):
            self.lp_trig_const[i] = self.double_press_speed / DOUBLE_PRESS_RATIO
            self.lp_cool_down_const[i] = self.double_press_speed

    def set_sensitivity(self):
        """
        Function to map the idea of "sensitivity between 0-1" to an actual threshold.
        """
        inv_sensitivity = (
            (SENSITIVITY_MAX - self.sensitivity)
        ) * 10.0  # since intuitively high sensitivity -> more likely to detect
        sens = 2 ** (1.1 * inv_sensitivity) * 600  # Yielding a nice exponential curve between 0-10
        self.threshold_trig = int(sens)
        self.threshold_cool_down = int(self.threshold_trig / THRESHOLD_RATIO)

    def lp_cool_down_filter(self, data, index):
        ret = int(
            self.lp_cool_down_const[index] * self.lp_cool_down[index]
            + (1.0 - self.lp_cool_down_const[index]) * data
        )
        self.lp_cool_down[index] = ret
        return ret

    def lp_trig_filter(self, data, index):
        ret = int(
            self.lp_trig_const[index] * self.lp_trig[index]
            + (1.0 - self.lp_trig_const[index]) * data
        )
        self.lp_trig[index] = ret
        return ret

    def lp_average_filter(self, data, index):
        ret = int(
            self.lp_average_const[index] * self.lp_average[index]
            + (1.0 - self.lp_average_const[index]) * data
        )
        self.lp_average[index] = ret
        return ret

    def calibrate(self, index):
        # Update the lp_average constant based on the noise levels.
        max_diff = max(
            abs(
                self.average_history[-self.calibration_limit :, index]
                - self.signal_history[-self.calibration_limit :, index]
            )
        )

        lp_new = (self.lp_average_const[index] * self.noise_level_target) / max_diff
        if lp_new < LP_AVERAGE_CONST_MAX:
            self.lp_average_const[index] = lp_new

    def detect(self, trig_val, cool_down_val, index):
        trig = False

        if trig_val > self.threshold_trig:
            if all(self.chilled):
                trig = True
            self.chilled[index] = False
        if cool_down_val < self.threshold_cool_down:
            self.chilled[index] = True

        return trig

    def process(self, frame, data_info):

        detections = [False] * self.num_depths
        signal = [0.0] * self.num_depths

        for i in range(frame.shape[1]):

            if not self.initialized[i]:
                self.lp_average[i] = int(np.mean(frame[:, i], axis=0))
                self.initialized[i] = True

            signal[i] = np.mean(frame[:, i], axis=0)
            self.signal_history[-1, -1] = max(self.signal_history[-1, :-1])

            lp_average = self.lp_average_filter(signal[i], i)  # Low pass to smooth the signal
            self.average_history[-1, -1] = max(self.average_history[-1, :-1])

            diff = int(abs(lp_average - signal[i]))

            if diff < self.noise_level_target:
                diff = 0  # Get back to "detect" state asap

            diff_sq = (
                diff * diff
            )  # Squaring yields a bit simpler behaviour with regards to thresholds.

            trig_val = self.lp_trig_filter(diff_sq, i)
            cool_down_val = self.lp_cool_down_filter(diff_sq, i)

            # Triggering checks if it is considered a button press.
            # Cool down checks if we can trigger again.
            detection = self.detect(trig_val, cool_down_val, i)

            if self.frame_count % self.recalibration_frame_period == self.calibration_limit:
                self.calibrate(i)
                self.calibrated = 5

            if self.frame_count <= self.calibration_limit:
                detection = False

            self.signal_history[-1, i] = signal[i]

            self.average_history[-1, i] = lp_average

            self.trig_history[-1, i] = trig_val

            self.cool_down_history[-1, i] = cool_down_val

            detections[i] = detection

        detection = any(detections)
        self.cool_down_history[-1, -1] = max(self.cool_down_history[-1, :-1])

        self.trig_history[-1, -1] = max(self.trig_history[-1, :-1])

        self.cool_down_history = np.roll(self.cool_down_history, -1, axis=0)
        self.trig_history = np.roll(self.trig_history, -1, axis=0)
        self.average_history = np.roll(self.average_history, -1, axis=0)
        self.signal_history = np.roll(self.signal_history, -1, axis=0)

        if detection:
            self.detection_history.append(self.frame_count)

        if self.calibrated > 0:
            calibrated = True
            self.calibrated -= 1
        else:
            calibrated = False

        out_data = {
            "signal_history": self.signal_history[:, -1],
            "average_history": self.average_history[:, -1],
            "trig_history": self.trig_history[:, -1],
            "cool_down_history": self.cool_down_history[:, -1],
            "threshold": self.threshold_trig,
            "detection_history": (np.array(self.detection_history) - self.frame_count) / self.f,
            "detection": detection,
            "signal": signal,
            "calibrated": calibrated,
        }

        self.frame_count += 1

        return out_data



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Tuple

import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121


def get_distances_m(
    config: a121.SensorConfig,
    metadata: a121.Metadata,
) -> Tuple[npt.NDArray[np.float_], float]:
    points = np.arange(config.num_points) * config.step_length + config.start_point
    distances_m = points * metadata.base_step_length_m
    step_length_m = config.step_length * metadata.base_step_length_m
    return distances_m, step_length_m


def get_approx_sweep_rate(config: a121.SensorConfig) -> float:
    ppp = {1: 24, 2: 20, 3: 16, 4: 16, 5: 16}[config.profile.value]

    n = 3 * ppp + config.num_points * config.hwaas * ppp

    return config.prf.frequency / n


def get_approx_fft_vels(config: a121.SensorConfig) -> Tuple[npt.NDArray, float]:
    sweep_rate = get_approx_sweep_rate(config)
    if config.sweep_rate is not None:
        sweep_rate = min([sweep_rate, config.sweep_rate])

    spf = config.sweeps_per_frame
    f_res = 1 / spf
    freqs = np.fft.fftshift(np.fft.fftfreq(spf))  # type: ignore[call-overload]
    f_to_v = 2.5e-3 * sweep_rate
    return freqs * f_to_v, f_res * f_to_v



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Any, Callable, Generic, Optional, Type

import attrs

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121 import algo
from acconeer.exptool.a121.algo._base import ConfigT, InputT, MetadataT, ResultT
from acconeer.exptool.app.new import GeneralMessage

from ._null_app_model import NullAppModel
from .processor import GenericProcessorPlotPluginBase


def processor_main(
    *,
    processor_cls: Type[algo.GenericProcessorBase[InputT, ConfigT, ResultT, MetadataT]],
    processor_config_cls: Type[ConfigT],
    plot_plugin: Type[GenericProcessorPlotPluginBase[ResultT, MetadataT]],
    sensor_config_getter: Callable[[], a121.SensorConfig],
) -> None:
    parser = a121.ExampleArgumentParser()
    parser.add_argument("--sensor", type=int, default=1)
    args = parser.parse_args()
    et.utils.config_logging(args)

    client = a121.Client(**a121.get_client_args(args))
    client.connect()

    sensor_config = sensor_config_getter()
    session_config = a121.SessionConfig({args.sensor: sensor_config})

    metadata = client.setup_session(session_config)

    processor_config = processor_config_cls()

    processor = processor_cls(
        sensor_config=sensor_config,
        metadata=metadata,  # type: ignore[arg-type]
        processor_config=processor_config,
    )

    pg_updater = ProcessorPGUpdater[ResultT, MetadataT](
        plot_plugin=plot_plugin,
        sensor_config=sensor_config,
        metadata=metadata,  # type: ignore[arg-type]
    )
    pg_process = et.PGProcess(pg_updater)  # type: ignore[attr-defined]
    pg_process.start()

    client.start_session()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        result = client.get_next()

        processor_result = processor.process(result)  # type: ignore[arg-type]

        try:
            pg_process.put_data(processor_result)
        except et.PGProccessDiedException:  # type: ignore[attr-defined]
            break

    print("Disconnecting...")
    pg_process.close()
    client.disconnect()


@attrs.mutable(kw_only=True, slots=False)
class ProcessorPGUpdater(Generic[ResultT, MetadataT]):
    plot_plugin: Type[GenericProcessorPlotPluginBase[ResultT, MetadataT]] = attrs.field()
    sensor_config: a121.SensorConfig = attrs.field()
    metadata: MetadataT = attrs.field()
    plot_plugin_obj: Optional[GenericProcessorPlotPluginBase[ResultT, MetadataT]] = attrs.field(
        default=None, init=False
    )

    def setup(self, win: pg.GraphicsLayout) -> None:
        self.plot_plugin_obj = self.plot_plugin(plot_layout=win, app_model=NullAppModel())
        self.plot_plugin_obj.handle_message(
            GeneralMessage(
                name="setup", kwargs=dict(sensor_config=self.sensor_config, metadata=self.metadata)
            )
        )

    def update(self, data: Any) -> None:
        if self.plot_plugin_obj is None:
            raise RuntimeError

        self.plot_plugin_obj.handle_message(GeneralMessage(name="plot", data=data))
        self.plot_plugin_obj.draw()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import abc
import logging
from typing import Generic, Type

import qtawesome as qta

from PySide6.QtWidgets import QPushButton, QVBoxLayout, QWidget

from acconeer.exptool import a121
from acconeer.exptool.a121.algo._base import ConfigT
from acconeer.exptool.a121.algo._plugins._a121 import A121ViewPluginBase
from acconeer.exptool.app.new import (
    BUTTON_ICON_COLOR,
    AppModel,
    ConnectionState,
    GeneralMessage,
    PluginState,
)
from acconeer.exptool.app.new.ui.plugin import (
    AttrsConfigEditor,
    GridGroupBox,
    PidgetFactoryMapping,
    SessionConfigEditor,
    SmartMetadataView,
    SmartPerfCalcView,
)

from .backend_plugin import ProcessorBackendPluginSharedState


log = logging.getLogger(__name__)


class ProcessorViewPluginBase(Generic[ConfigT], A121ViewPluginBase):
    def __init__(self, view_widget: QWidget, app_model: AppModel) -> None:
        super().__init__(app_model=app_model, view_widget=view_widget)
        self.layout = QVBoxLayout(self.view_widget)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.view_widget.setLayout(self.layout)

        self.start_button = QPushButton(
            qta.icon("fa5s.play-circle", color=BUTTON_ICON_COLOR),
            "Start measurement",
            self.view_widget,
        )
        self.stop_button = QPushButton(
            qta.icon("fa5s.stop-circle", color=BUTTON_ICON_COLOR),
            "Stop",
            self.view_widget,
        )
        self.defaults_button = QPushButton(
            qta.icon("mdi6.restore", color=BUTTON_ICON_COLOR),
            "Restore default settings",
            self.view_widget,
        )
        self.start_button.clicked.connect(self._send_start_requests)
        self.stop_button.clicked.connect(self._send_stop_requests)
        self.defaults_button.clicked.connect(self._send_defaults_request)

        self.start_button.setShortcut("space")
        self.stop_button.setShortcut("space")

        button_group = GridGroupBox("Controls", parent=self.view_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)
        button_group.layout().addWidget(self.defaults_button, 1, 0, 1, 2)

        self.layout.addWidget(button_group)

        self.metadata_view = SmartMetadataView(self.view_widget)
        self.layout.addWidget(self.metadata_view)

        self.perf_calc_view = SmartPerfCalcView(self.view_widget)
        self.layout.addWidget(self.perf_calc_view)

        self.session_config_editor = SessionConfigEditor(self.view_widget)
        self.session_config_editor.sig_update.connect(self._on_session_config_update)
        self.processor_config_editor = AttrsConfigEditor[ConfigT](
            title="Processor parameters",
            factory_mapping=self.get_pidget_mapping(),
            parent=self.view_widget,
        )
        self.processor_config_editor.sig_update.connect(self._on_processor_config_update)
        self.layout.addWidget(self.processor_config_editor)
        self.layout.addWidget(self.session_config_editor)
        self.layout.addStretch()

    def _on_session_config_update(self, session_config: a121.SessionConfig) -> None:
        self.app_model.put_backend_plugin_task(
            "update_session_config", {"session_config": session_config}
        )

    def _on_processor_config_update(self, processor_config: ConfigT) -> None:
        self.app_model.put_backend_plugin_task(
            "update_processor_config", {"processor_config": processor_config}
        )

    def _send_start_requests(self) -> None:
        self.app_model.put_backend_plugin_task("start_session", on_error=self.app_model.emit_error)
        self.app_model.set_plugin_state(PluginState.LOADED_STARTING)

    def _send_stop_requests(self) -> None:
        self.app_model.put_backend_plugin_task("stop_session", on_error=self.app_model.emit_error)
        self.app_model.set_plugin_state(PluginState.LOADED_STOPPING)

    def _send_defaults_request(self) -> None:
        self.app_model.put_backend_plugin_task("restore_defaults")

    def teardown(self) -> None:
        self.layout.deleteLater()

    def handle_message(self, message: GeneralMessage) -> None:
        if message.name == "sync":
            log.debug(f"{type(self).__name__} syncing")

            self.session_config_editor.sync()
            self.processor_config_editor.sync()
        else:
            raise RuntimeError("Unknown message")

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.session_config_editor.setEnabled(app_model.plugin_state == PluginState.LOADED_IDLE)
        self.processor_config_editor.setEnabled(app_model.plugin_state == PluginState.LOADED_IDLE)
        self.start_button.setEnabled(
            app_model.plugin_state == PluginState.LOADED_IDLE
            and app_model.connection_state == ConnectionState.CONNECTED
            and app_model.backend_plugin_state.ready
        )
        self.stop_button.setEnabled(app_model.plugin_state == PluginState.LOADED_BUSY)
        self.defaults_button.setEnabled(app_model.plugin_state == PluginState.LOADED_IDLE)

        if app_model.backend_plugin_state is None:
            self.session_config_editor.set_data(None)
            self.processor_config_editor.set_data(None)
            self.metadata_view.update(None)
            self.perf_calc_view.update()
        else:
            assert isinstance(app_model.backend_plugin_state, ProcessorBackendPluginSharedState)
            assert isinstance(
                app_model.backend_plugin_state.processor_config, self.get_processor_config_cls()
            )

            self.session_config_editor.set_data(app_model.backend_plugin_state.session_config)
            self.processor_config_editor.set_data(app_model.backend_plugin_state.processor_config)
            self.metadata_view.update(app_model.backend_plugin_state.metadata)
            self.perf_calc_view.update(
                app_model.backend_plugin_state.session_config,
                app_model.backend_plugin_state.metadata,
            )

        self.session_config_editor.update_available_sensor_list(app_model._a121_server_info)

    @classmethod
    @abc.abstractmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        pass

    @classmethod
    @abc.abstractmethod
    def get_processor_config_cls(cls) -> Type[ConfigT]:
        pass



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import platform
from typing import Optional, Type, Union

import attrs

import acconeer.exptool as et
from acconeer.exptool.a121._core.entities import ClientInfo
from acconeer.exptool.a121._core.mediators import (
    AgnosticClient,
    BufferedLink,
    ClientError,
    CommunicationProtocol,
)
from acconeer.exptool.utils import USBDevice  # type: ignore[import]

from .exploration_protocol import ExplorationProtocol, get_exploration_protocol
from .links import AdaptedSerialLink, AdaptedSocketLink, AdaptedUSBLink, NullLink, NullLinkError


def determine_serial_port(serial_port: Optional[str]) -> str:
    if serial_port is None:
        tagged_ports = et.utils.get_tagged_serial_ports()
        A121_TAGS = ["XC120"]
        hopefully_a_single_tagged_port = [port for port, tag in tagged_ports if tag in A121_TAGS]
        try:
            (port,) = hopefully_a_single_tagged_port
            return str(port)
        except ValueError:
            if hopefully_a_single_tagged_port == []:
                raise ClientError("No serial devices detected. Cannot auto detect.")
            else:
                port_list = "\n".join(
                    [f"* {port} ({tag})" for port, tag in hopefully_a_single_tagged_port]
                )

                raise ClientError(
                    "There are multiple devices detected. Specify one:\n" + port_list
                )
    else:
        return serial_port


def determine_usb_device(usb_device: Optional[USBDevice]) -> USBDevice:
    if usb_device is None:
        usb_devices = et.utils.get_usb_devices()

        if not usb_devices:
            raise ClientError("No USB devices detected. Cannot auto detect.")
        elif len(usb_devices) > 1:
            raise ClientError("There are multiple devices detected. Specify one:\n" + usb_devices)
        else:
            return usb_devices[0]
    else:
        return usb_device


def link_factory(client_info: ClientInfo) -> BufferedLink:

    if client_info.ip_address is not None:
        return AdaptedSocketLink(host=client_info.ip_address)

    if client_info.serial_port is not None:
        link = AdaptedSerialLink(
            port=client_info.serial_port,
        )
        if client_info.override_baudrate is not None:
            link.baudrate = client_info.override_baudrate

        return link

    if client_info.usb_device is not None:
        link = AdaptedUSBLink(
            vid=client_info.usb_device.vid,
            pid=client_info.usb_device.pid,
            name=client_info.usb_device.name,
        )

        return link

    return NullLink()


def autodetermine_client_link(client_info: ClientInfo) -> ClientInfo:
    if platform.system().lower() == "windows":
        try:
            client_info = attrs.evolve(
                client_info,
                usb_device=determine_usb_device(client_info.usb_device),
            )

            return client_info
        except ClientError:
            pass
    try:
        client_info = attrs.evolve(
            client_info,
            serial_port=determine_serial_port(client_info.serial_port),
        )
        return client_info
    except ClientError:
        raise ClientError("No devices detected. Cannot auto detect.")


class Client(AgnosticClient):
    _protocol_overridden: bool
    _client_info: ClientInfo

    def __init__(
        self,
        ip_address: Optional[str] = None,
        serial_port: Optional[str] = None,
        usb_device: Optional[Union[str, et.utils.USBDevice]] = None,
        override_baudrate: Optional[int] = None,
        _override_protocol: Optional[Type[CommunicationProtocol]] = None,
    ):
        if len([e for e in [ip_address, serial_port, usb_device] if e is not None]) > 1:
            raise ValueError("Only one connection can be selected")

        if isinstance(usb_device, str):
            raise NotImplementedError("Selecting device by serial number not supported")

        protocol: Type[CommunicationProtocol] = ExplorationProtocol
        self._protocol_overridden = False

        if _override_protocol is not None:
            protocol = _override_protocol
            self._protocol_overridden = True

        self._client_info = ClientInfo(
            ip_address=ip_address,
            override_baudrate=override_baudrate,
            serial_port=serial_port,
            usb_device=usb_device,
        )

        super().__init__(
            link=link_factory(self._client_info),
            protocol=protocol,
        )

    @property
    def client_info(self) -> ClientInfo:
        return self._client_info

    def connect(self) -> None:
        try:
            super().connect()
        except NullLinkError:
            self._client_info = autodetermine_client_link(self._client_info)
            self._link = link_factory(self.client_info)
            super().connect()

        if not self._protocol_overridden:
            if issubclass(self._protocol, ExplorationProtocol):
                try:
                    new_protocol = get_exploration_protocol(self.server_info.parsed_rss_version)
                except Exception:
                    self.disconnect()
                    raise
                else:
                    self._protocol = new_protocol



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import importlib.resources
import sys

import qdarktheme

from PySide6 import QtCore, QtGui
from PySide6.QtWidgets import QApplication

import pyqtgraph as pg

from acconeer.exptool.app import resources  # type: ignore[attr-defined]
from acconeer.exptool.utils import config_logging  # type: ignore[import]

from ._argument_parser import ExptoolArgumentParser
from .app_model import AppModel
from .backend import Backend
from .plugin_loader import load_default_plugins
from .storage import remove_config_dir, remove_temp_dir
from .ui import AppModelViewer, MainWindow


def main():
    parser = ExptoolArgumentParser()
    args = parser.parse_args()
    config_logging(args)

    if args.purge_config:
        remove_config_dir()
        remove_temp_dir()
        print("Config purged")
        sys.exit(0)

    backend = Backend()
    backend.start()

    model = AppModel(backend, load_default_plugins())
    model.start()

    pg.setConfigOption("background", "w")
    pg.setConfigOption("foreground", "k")
    pg.setConfigOption("leftButtonPan", False)
    pg.setConfigOptions(antialias=True)

    QApplication.setHighDpiScaleFactorRoundingPolicy(
        QtCore.Qt.HighDpiScaleFactorRoundingPolicy.PassThrough,
    )
    QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)

    app = QApplication.instance()
    if app is None:
        app = QApplication(sys.argv)

    app.setStyleSheet(qdarktheme.load_stylesheet("light"))
    app.setAttribute(QtCore.Qt.ApplicationAttribute.AA_UseHighDpiPixmaps)

    with importlib.resources.path(resources, "icon.png") as path:
        app.setWindowIcon(_pixmap_to_icon(QtGui.QPixmap(str(path))))

    mw = MainWindow(model)
    mw.show()

    if args.amv:
        app_model_viewer = AppModelViewer(model)
        app_model_viewer.show()

    model.broadcast()

    app.exec()

    model.stop()
    backend.stop()


def _pixmap_to_icon(pixmap: QtGui.QPixmap) -> QtGui.QIcon:
    size = max(pixmap.size().height(), pixmap.size().width())

    square_pixmap = QtGui.QPixmap(size, size)
    square_pixmap.fill(QtGui.Qt.transparent)

    painter = QtGui.QPainter(square_pixmap)
    painter.drawPixmap(
        (square_pixmap.size().width() - pixmap.size().width()) // 2,
        (square_pixmap.size().height() - pixmap.size().height()) // 2,
        pixmap,
    )
    painter.end()

    scaled_pixmap = square_pixmap.scaled(
        256,
        256,
        aspectMode=QtGui.Qt.KeepAspectRatio,
        mode=QtGui.Qt.SmoothTransformation,
    )

    return QtGui.QIcon(scaled_pixmap)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import logging
import time
from typing import Any, Optional, Tuple, Type, Union

import attrs

from acconeer.exptool.a121._core.entities import (
    ClientInfo,
    Metadata,
    Result,
    SensorConfig,
    ServerInfo,
    SessionConfig,
)
from acconeer.exptool.a121._core.utils import (
    create_extended_structure,
    iterate_extended_structure,
    unextend,
    unwrap_ticks,
)
from acconeer.exptool.a121._perf_calc import _PerformanceCalc

from .communication_protocol import CommunicationProtocol
from .link import BufferedLink
from .recorder import Recorder


log = logging.getLogger(__name__)


class ClientError(Exception):
    pass


class AgnosticClient:
    _link: BufferedLink
    _default_link_timeout: float
    _link_timeout: float
    _protocol: Type[CommunicationProtocol]
    _server_info: Optional[ServerInfo]
    _session_config: Optional[SessionConfig]
    _metadata: Optional[list[dict[int, Metadata]]]
    _session_is_started: bool
    _recorder: Optional[Recorder]
    _tick_unwrapper: TickUnwrapper

    def __init__(self, link: BufferedLink, protocol: Type[CommunicationProtocol]) -> None:
        self._link = link
        self._protocol = protocol
        self._server_info = None
        self._session_config = None
        self._session_is_started = False
        self._metadata = None
        self._recorder = None
        self._tick_unwrapper = TickUnwrapper()

    def _assert_connected(self):
        if not self.connected:
            raise ClientError("Client is not connected.")

    def _assert_session_setup(self):
        self._assert_connected()
        if not self.session_is_setup:
            raise ClientError("Session is not set up.")

    def _assert_session_started(self):
        self._assert_session_setup()
        if not self.session_is_started:
            raise ClientError("Session is not started.")

    def connect(self) -> None:
        """Connects to the specified host.

        :raises: Exception if the host cannot be connected to.
        """
        self._default_link_timeout = self._link.timeout
        self._link_timeout = self._default_link_timeout
        self._link.connect()

        self._link.send(self._protocol.get_sensor_info_command())
        sens_response = self._link.recv_until(self._protocol.end_sequence)
        sensor_infos = self._protocol.get_sensor_info_response(sens_response)

        self._link.send(self._protocol.get_system_info_command())
        sys_response = self._link.recv_until(self._protocol.end_sequence)
        self._server_info, sensor = self._protocol.get_system_info_response(
            sys_response, sensor_infos
        )

        if sensor != "a121":
            self._link.disconnect()
            raise ClientError(f"Wrong sensor version, expected a121 but got {sensor}")

    def setup_session(
        self,
        config: Union[SensorConfig, SessionConfig],
    ) -> Union[Metadata, list[dict[int, Metadata]]]:
        """Sets up the session specified by ``config``.

        If the Client is not already connected, it will connect before setting up the session.

        :param config: The session to set up.
        :raises:
            ``ValueError`` if the config is invalid.

        :returns:
            ``Metadata`` if ``config.extended is False``,
            ``list[dict[int, Metadata]]`` otherwise.
        """
        if not self.connected:
            self.connect()

        if self.session_is_started:
            raise ClientError("Session is currently running, can't setup.")

        if isinstance(config, SensorConfig):
            config = SessionConfig(config)

        config.validate()

        pc = _PerformanceCalc(config, None)

        try:
            # Increase timeout if update rate is very low, otherwise keep default
            self._link_timeout = max(1.1 * (1 / pc.frame_rate) + 0.5, self._link.timeout)
        except Exception:
            self._link_timeout = self._default_link_timeout

        self._link.timeout = self._link_timeout

        self._link.send(self._protocol.setup_command(config))
        reponse_bytes = self._link.recv_until(self._protocol.end_sequence)
        self._session_config = config
        self._metadata = self._protocol.setup_response(
            reponse_bytes, context_session_config=config
        )

        if self.session_config.extended:
            return self._metadata
        else:
            return unextend(self._metadata)

    def start_session(self, recorder: Optional[Recorder] = None) -> None:
        """Starts the already set up session.

        After this call, the server starts streaming data to the client.

        :param recorder:
            An optional ``Recorder``, which samples every ``get_next()``
        :raises: ``ClientError`` if ``Client``'s  session is not set up.
        """
        self._assert_session_setup()

        if self.session_is_started:
            raise ClientError("Session is already started.")

        if recorder is not None:
            self._recorder = recorder
            self._recorder._start(
                client_info=self.client_info,
                extended_metadata=self.extended_metadata,
                server_info=self.server_info,
                session_config=self.session_config,
            )

        self._link.timeout = self._link_timeout

        self._link.send(self._protocol.start_streaming_command())
        reponse_bytes = self._link.recv_until(self._protocol.end_sequence)
        self._protocol.start_streaming_response(reponse_bytes)
        self._session_is_started = True

    def get_next(self) -> Union[Result, list[dict[int, Result]]]:
        """Gets results from the server.

        :returns:
            A ``Result`` if the setup ``SessionConfig.extended is False``,
            ``list[dict[int, Result]]`` otherwise.
        :raises:
            ``ClientError`` if ``Client``'s session is not started.
        """
        self._assert_session_started()

        payload_size, partial_results = self._protocol.get_next_header(
            bytes_=self._link.recv_until(self._protocol.end_sequence),
            extended_metadata=self.extended_metadata,
            ticks_per_second=self.server_info.ticks_per_second,
        )
        payload = self._link.recv(payload_size)
        extended_results = self._protocol.get_next_payload(payload, partial_results)

        extended_results = self._tick_unwrapper.unwrap_ticks(extended_results)

        if self._recorder is not None:
            self._recorder._sample(extended_results)

        if self.session_config.extended:
            return extended_results
        else:
            return unextend(extended_results)

    def stop_session(self) -> Any:
        """Stops an on-going session

        :returns:
            The return value of the passed ``Recorder.stop()`` passed in ``start_session``.
        :raises:
            ``ClientError`` if ``Client``'s session is not started.
        """
        self._assert_session_started()

        recorder_result = None
        if self._recorder is not None:
            recorder_result = self._recorder._stop()
            self._recorder = None

        try:
            self._link.send(self._protocol.stop_streaming_command())
            reponse_bytes = self._drain_buffer(self._link.timeout + 1)
            self._protocol.stop_streaming_response(reponse_bytes)
        except Exception:
            raise
        finally:
            self._link.timeout = self._default_link_timeout
            self._session_is_started = False
            self._tick_unwrapper = TickUnwrapper()

        return recorder_result

    def _drain_buffer(
        self, timeout_s: float = 3.0
    ) -> bytes:  # TODO: Make `timeout_s` session-dependant
        """Drains data in the buffer. Returning the first bytes that are not data packets."""
        start = time.time()

        while time.time() < start + timeout_s:
            next_header = self._link.recv_until(self._protocol.end_sequence)
            try:
                payload_size, _ = self._protocol.get_next_header(
                    bytes_=next_header,
                    extended_metadata=self.extended_metadata,
                    ticks_per_second=self.server_info.ticks_per_second,
                )
                _ = self._link.recv(payload_size)
            except Exception:
                return next_header
            else:
                log.debug("Threw away get_next package when draining buffer")
        raise ClientError("Client timed out when waiting for 'stop'-response.")

    def disconnect(self) -> None:
        """Disconnects the client from the host.

        :raises: ``ClientError`` if ``Client`` is not connected.
        """
        # TODO: Make sure this cleans up corner-cases (like lost connection)
        #       to not hog resources.
        self._assert_connected()

        if self.session_is_started:
            _ = self.stop_session()

        self._server_info = None
        self._link.disconnect()

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, type_, value, traceback):
        self.disconnect()

    @property
    def connected(self) -> bool:
        """Whether this Client is connected."""
        return self._server_info is not None

    @property
    def session_is_setup(self) -> bool:
        """Whether this Client has a session set up."""
        return self._session_config is not None

    @property
    def session_is_started(self) -> bool:
        """Whether this Client's session is started."""
        return self._session_is_started

    @property
    def server_info(self) -> ServerInfo:
        """The ``ServerInfo``."""
        self._assert_connected()

        return self._server_info  # type: ignore[return-value]

    @property
    def client_info(self) -> ClientInfo:
        """The ``ClientInfo``."""
        return ClientInfo()

    @property
    def session_config(self) -> SessionConfig:
        """The :class:`SessionConfig` for the current session"""

        self._assert_session_setup()
        assert self._session_config is not None  # Should never happen if session is setup
        return self._session_config

    @property
    def extended_metadata(self) -> list[dict[int, Metadata]]:
        """The extended :class:`Metadata` for the current session"""

        self._assert_session_setup()
        assert self._metadata is not None  # Should never happen if session is setup
        return self._metadata


class TickUnwrapper:
    """Wraps unwrap_ticks to be applied over extended results"""

    def __init__(self) -> None:
        self.next_minimum_tick: Optional[int] = None

    def unwrap_ticks(self, extended_results: list[dict[int, Result]]) -> list[dict[int, Result]]:
        result_items = list(iterate_extended_structure(extended_results))
        ticks = [result.tick for _, _, result in result_items]
        unwrapped_ticks, self.next_minimum_tick = unwrap_ticks(ticks, self.next_minimum_tick)

        def f(result_item: Tuple[int, int, Result], updated_tick: int) -> Tuple[int, int, Result]:
            group_index, sensor_id, result = result_item
            updated_result = attrs.evolve(result, tick=updated_tick)
            return (group_index, sensor_id, updated_result)

        return create_extended_structure(map(f, result_items, unwrapped_ticks))



import ctypes
import struct
from ctypes import (
    byref,
    c_byte,
    c_ubyte,
    c_ulong,
    c_void_p,
    create_string_buffer,
    resize,
    sizeof,
    wstring_at,
)
from ctypes.wintypes import DWORD

from .winusb import WinUSBApi
from .winusbclasses import (
    DIGCF_ALLCLASSES,
    DIGCF_DEFAULT,
    DIGCF_DEVICE_INTERFACE,
    DIGCF_PRESENT,
    DIGCF_PROFILE,
    ERROR_IO_INCOMPLETE,
    ERROR_IO_PENDING,
    FILE_ATTRIBUTE_NORMAL,
    FILE_FLAG_OVERLAPPED,
    FILE_SHARE_READ,
    FILE_SHARE_WRITE,
    GENERIC_READ,
    GENERIC_WRITE,
    GUID,
    INVALID_HANDLE_VALUE,
    OPEN_EXISTING,
    Overlapped,
    PipeInfo,
    SpDeviceInterfaceData,
    SpDeviceInterfaceDetailData,
    SpDevinfoData,
    UsbInterfaceDescriptor,
)
from .winusbutils import (
    SPDRP_FRIENDLYNAME,
    Close_Handle,
    CreateFile,
    GetLastError,
    SetupDiEnumDeviceInterfaces,
    SetupDiGetClassDevs,
    SetupDiGetDeviceInterfaceDetail,
    SetupDiGetDeviceRegistryProperty,
    WinUsb_ControlTransfer,
    WinUsb_FlushPipe,
    WinUsb_Free,
    WinUsb_GetAssociatedInterface,
    WinUsb_GetOverlappedResult,
    WinUsb_Initialize,
    WinUsb_QueryDeviceInformation,
    WinUsb_QueryInterfaceSettings,
    WinUsb_QueryPipe,
    WinUsb_ReadPipe,
    WinUsb_SetPipePolicy,
    WinUsb_WritePipe,
    is_device,
)


def is_64bit():
    return struct.calcsize("P") * 8 == 64


class WinUsbPy(object):
    def __init__(self):
        self.api = WinUSBApi()
        byte_array = c_byte * 8
        self.usb_device_guid = GUID(
            0xA5DCBF10, 0x6530, 0x11D2, byte_array(0x90, 0x1F, 0x00, 0xC0, 0x4F, 0xB9, 0x51, 0xED)
        )
        self.usb_winusb_guid = GUID(
            0xDEE824EF, 0x729B, 0x4A0E, byte_array(0x9C, 0x14, 0xB7, 0x11, 0x7D, 0x33, 0xA8, 0x17)
        )
        self.usb_composite_guid = GUID(
            0x36FC9E60, 0xC465, 0x11CF, byte_array(0x80, 0x56, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00)
        )
        self.handle_file = INVALID_HANDLE_VALUE
        self.handle_winusb = [c_void_p()]
        self._index = -1
        self.device_paths = {}  # type: dict

    def list_usb_devices(
        self,
        default=False,
        present=False,
        allclasses=False,
        profile=False,
        deviceinterface=False,
        vid=None,
        pid=None,
        name=None,
    ):
        self.device_paths = {}
        value = 0x00000000
        try:
            if default:
                value |= DIGCF_DEFAULT
            if present:
                value |= DIGCF_PRESENT
            if allclasses:
                value |= DIGCF_ALLCLASSES
            if profile:
                value |= DIGCF_PROFILE
            if deviceinterface:
                value |= DIGCF_DEVICE_INTERFACE

        except KeyError:
            if value == 0x00000000:
                value = 0x00000010
            pass

        flags = DWORD(value)
        self.handle = self.api.exec_function_setupapi(
            SetupDiGetClassDevs, byref(self.usb_winusb_guid), None, None, flags
        )

        sp_device_interface_data = SpDeviceInterfaceData()
        sp_device_interface_data.cb_size = sizeof(sp_device_interface_data)
        sp_device_interface_detail_data = SpDeviceInterfaceDetailData()
        sp_device_info_data = SpDevinfoData()
        sp_device_info_data.cb_size = sizeof(sp_device_info_data)

        i = 0
        required_size = DWORD(0)
        member_index = DWORD(i)
        cb_sizes = (8, 6, 5)  # different on 64 bit / 32 bit etc

        while self.api.exec_function_setupapi(
            SetupDiEnumDeviceInterfaces,
            self.handle,
            None,
            byref(self.usb_winusb_guid),
            member_index,
            byref(sp_device_interface_data),
        ):
            self.api.exec_function_setupapi(
                SetupDiGetDeviceInterfaceDetail,
                self.handle,
                byref(sp_device_interface_data),
                None,
                0,
                byref(required_size),
                None,
            )
            resize(sp_device_interface_detail_data, required_size.value)

            path = None
            for cb_size in cb_sizes:
                sp_device_interface_detail_data.cb_size = cb_size
                ret = self.api.exec_function_setupapi(
                    SetupDiGetDeviceInterfaceDetail,
                    self.handle,
                    byref(sp_device_interface_data),
                    byref(sp_device_interface_detail_data),
                    required_size,
                    byref(required_size),
                    byref(sp_device_info_data),
                )
                if ret:
                    cb_sizes = (cb_size,)
                    path = wstring_at(byref(sp_device_interface_detail_data, sizeof(DWORD)))
                    break
            if path is None:
                raise ctypes.WinError()

            if vid is not None and pid is not None:
                if is_device(name, vid, pid, path):
                    self.device_paths[path] = path
                    return self.device_paths
            else:

                # friendly name
                uname = path
                buff_friendly_name = ctypes.create_unicode_buffer(250)
                if self.api.exec_function_setupapi(
                    SetupDiGetDeviceRegistryProperty,
                    self.handle,
                    byref(sp_device_info_data),
                    SPDRP_FRIENDLYNAME,
                    None,
                    ctypes.byref(buff_friendly_name),
                    ctypes.sizeof(buff_friendly_name) - 1,
                    None,
                ):

                    uname = buff_friendly_name.value
                # else:
                #     err = self.get_last_error_code()
                #     # print(ctypes.WinError())
                self.device_paths[uname] = path
                if name is not None and uname == name:
                    return self.device_paths
            i += 1
            member_index = DWORD(i)
            required_size = c_ulong(0)
            resize(sp_device_interface_detail_data, sizeof(SpDeviceInterfaceDetailData))
        return self.device_paths

    def find_device(self, path):
        return is_device(self._name, self._vid, self._pid, path)

    def init_winusb_device(self, name, vid, pid):
        self._vid = vid
        self._pid = pid
        self._name = name
        path = None
        try:
            for name, p in self.device_paths.items():
                if self.find_device(name) or self.find_device(p):
                    path = p
                    break

        except IndexError:
            return False

        if path is None:
            return False

        self.handle_file = self.api.exec_function_kernel32(
            CreateFile,
            path,
            GENERIC_WRITE | GENERIC_READ,
            FILE_SHARE_WRITE | FILE_SHARE_READ,
            None,
            OPEN_EXISTING,
            FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED,
            None,
        )

        if self.handle_file == INVALID_HANDLE_VALUE:
            return False
        result = self.api.exec_function_winusb(
            WinUsb_Initialize, self.handle_file, byref(self.handle_winusb[0])
        )
        if result == 0:
            # err = self.get_last_error_code()
            raise ctypes.WinError()
            # return False
        else:
            self._index = 0
            return True

    def close_winusb_device(self):
        result_file = 1
        if self.handle_file:
            result_file = self.api.exec_function_kernel32(Close_Handle, self.handle_file)
            if result_file:
                self.handle_file = None

        result_winusb = [self.api.exec_function_winusb(WinUsb_Free, h) for h in self.handle_winusb]
        if 0 in result_winusb:
            raise RuntimeError("Unable to close winusb handle")
        self.handle_winusb = []
        return result_file != 0

    def get_last_error_code(self):
        return self.api.exec_function_kernel32(GetLastError)

    def query_device_info(self, query=1):
        info_type = c_ulong(query)
        buff = (c_void_p * 1)()
        buff_length = c_ulong(sizeof(c_void_p))
        result = self.api.exec_function_winusb(
            WinUsb_QueryDeviceInformation,
            self.handle_winusb[self._index],
            info_type,
            byref(buff_length),
            buff,
        )
        if result != 0:
            return buff[0]
        else:
            return -1

    def query_interface_settings(self, index):
        if self._index != -1:
            temp_handle_winusb = self.handle_winusb[self._index]
            interface_descriptor = UsbInterfaceDescriptor()
            result = self.api.exec_function_winusb(
                WinUsb_QueryInterfaceSettings,
                temp_handle_winusb,
                c_ubyte(0),
                byref(interface_descriptor),
            )
            if result != 0:
                return interface_descriptor
            else:
                return None
        else:
            return None

    def change_interface(self, index, alternate=0):
        new_handle = c_void_p()
        result = self.api.exec_function_winusb(
            WinUsb_GetAssociatedInterface,
            self.handle_winusb[self._index],
            c_ubyte(alternate),
            byref(new_handle),
        )
        if result != 0:
            self._index = index + 1
            self.handle_winusb.append(new_handle)
            return True
        else:
            return False

    def query_pipe(self, pipe_index):
        pipe_info = PipeInfo()
        result = self.api.exec_function_winusb(
            WinUsb_QueryPipe,
            self.handle_winusb[self._index],
            c_ubyte(0),
            pipe_index,
            byref(pipe_info),
        )
        if result != 0:
            return pipe_info
        else:
            return None

    def control_transfer(self, setup_packet, buff=None):
        if buff is not None:
            if setup_packet.length > 0:  # Host 2 Device
                buff = (c_ubyte * setup_packet.length)(*buff)
                buffer_length = setup_packet.length
            else:  # Device 2 Host
                buff = (c_ubyte * setup_packet.length)()
                buffer_length = setup_packet.length
        else:
            buff = c_ubyte()
            buffer_length = 0

        result = self.api.exec_function_winusb(
            WinUsb_ControlTransfer,
            self.handle_winusb[0],
            setup_packet,
            byref(buff),
            c_ulong(buffer_length),
            byref(c_ulong(0)),
            None,
        )
        return {"result": result != 0, "buffer": [buff]}

    def write(self, pipe_id, write_buffer):
        write_buffer = create_string_buffer(write_buffer)
        written = c_ulong(0)
        self.api.exec_function_winusb(
            WinUsb_WritePipe,
            self.handle_winusb[self._index],
            c_ubyte(pipe_id),
            write_buffer,
            c_ulong(len(write_buffer) - 1),
            byref(written),
            None,
        )
        return written.value

    def read(self, pipe_id, length_buffer):
        read_buffer = create_string_buffer(length_buffer)
        read = c_ulong(0)
        result = self.api.exec_function_winusb(
            WinUsb_ReadPipe,
            self.handle_winusb[self._index],
            c_ubyte(pipe_id),
            read_buffer,
            c_ulong(length_buffer),
            byref(read),
            None,
        )
        if result != 0:
            if read.value != length_buffer:
                return read_buffer[: read.value]
            else:
                return read_buffer
        else:
            return None

    def set_timeout(self, pipe_id, timeout):
        class POLICY_TYPE:
            SHORT_PACKET_TERMINATE = 1
            AUTO_CLEAR_STALL = 2
            PIPE_TRANSFER_TIMEOUT = 3
            IGNORE_SHORT_PACKETS = 4
            ALLOW_PARTIAL_READS = 5
            AUTO_FLUSH = 6
            RAW_IO = 7

        policy_type = c_ulong(POLICY_TYPE.PIPE_TRANSFER_TIMEOUT)
        value_length = c_ulong(4)
        value = c_ulong(int(timeout * 1000))  # in ms
        result = self.api.exec_function_winusb(
            WinUsb_SetPipePolicy,
            self.handle_winusb[self._index],
            c_ubyte(pipe_id),
            policy_type,
            value_length,
            byref(value),
        )
        return result

    def flush(self, pipe_id):
        result = self.api.exec_function_winusb(
            WinUsb_FlushPipe, self.handle_winusb[self._index], c_ubyte(pipe_id)
        )
        return result

    def _overlapped_read_do(self, pipe_id):
        self.olread_ol.Internal = 0
        self.olread_ol.InternalHigh = 0
        self.olread_ol.Offset = 0
        self.olread_ol.OffsetHigh = 0
        self.olread_ol.Pointer = 0
        self.olread_ol.hEvent = 0
        result = self.api.exec_function_winusb(
            WinUsb_ReadPipe,
            self.handle_winusb[self._index],
            c_ubyte(pipe_id),
            self.olread_buf,
            c_ulong(self.olread_buflen),
            byref(c_ulong(0)),
            byref(self.olread_ol),
        )
        if result != 0:
            return True
        else:
            return False

    def overlapped_read_init(self, pipe_id, length_buffer):
        self.olread_ol = Overlapped()
        self.olread_buf = create_string_buffer(length_buffer)
        self.olread_buflen = length_buffer
        return self._overlapped_read_do(pipe_id)

    def overlapped_read(self, pipe_id):
        """keep on reading overlapped, return bytearray, empty if nothing to read, None if err"""
        rl = c_ulong(0)
        result = self.api.exec_function_winusb(
            WinUsb_GetOverlappedResult,
            self.handle_winusb[self._index],
            byref(self.olread_ol),
            byref(rl),
            False,
        )
        if result == 0:
            if (
                self.get_last_error_code() == ERROR_IO_PENDING
                or self.get_last_error_code() == ERROR_IO_INCOMPLETE
            ):
                return ""
            else:
                return None
        else:
            ret = str(self.olread_buf[0 : rl.value])
            self._overlapped_read_do(pipe_id)
            return ret



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import abc
import argparse
import csv
import json
import os
from typing import Union

import numpy as np

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121 import _core, algo


try:
    import prettyprinter

    prettyprinter.install_extras(["attrs"])

    pprint = prettyprinter.cpprint
except ImportError:
    from pprint import pprint


DESCRIPTION = """This is a command line utility that lets you convert
.h5/.npz files to .csv-files for use as is or in
e.g. Microsoft Excel.

example usage:
  python3 convert_to_csv.py -v ~/my_data_file.h5 ~/my_output_file.csv
"""


class ConvertToCsvArgumentParser(argparse.ArgumentParser):
    def __init__(self) -> None:
        super().__init__(description=DESCRIPTION, formatter_class=argparse.RawTextHelpFormatter)
        self.add_argument(
            "-d",
            "--delimiter",
            choices=["c", "t"],
            dest="delimiter",
            default="c",
            help="Delimiter for the output data. Default is comma. 't' is for tab, 'c' for comma",
        )
        self.add_argument(
            "input_file",
            help='The input file with file endings ".h5" or ".npz" (only A111).',
        )
        self.add_argument(
            "output_file",
            help="The output file to which csv-data will be written.",
        )
        self.add_argument(
            "--index",
            "--id",
            "--sensor",
            metavar="index/id",
            dest="sensor",
            type=int,
            default=argparse.SUPPRESS,
            help="The sensor index. Gets data from a specific sensor when multiple sensors are "
            "used.",
        )
        self.add_argument(
            "-f",
            "--force",
            action="store_true",
            default=False,
            help='Forcefully overwrite "output_file" if it already exists.',
        )
        self.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            default=False,
            help='Prints meta data from "input_file".',
        )
        self.add_argument(
            "--sweep-as-column",
            action="store_true",
            default=False,
            help="Stores sweeps as columns instead of rows.\n"
            "The default is to store sweeps as rows.",
        )
        self.add_argument(
            "-m",
            "--add_sweep_metadata",
            action="store_true",
            default=False,
            help="Adds depth and sweep number info to the csv file",
        )


class TableConverter:
    @abc.abstractmethod
    def convert(self, sensor: int) -> np.ndarray:
        pass

    @abc.abstractmethod
    def get_metadata_rows(self, sensor: int) -> list[np.ndarray]:
        pass

    @abc.abstractmethod
    def print_information(self, verbose: bool = False) -> None:
        pass

    @staticmethod
    def format_cell_value(v):
        if np.imag(v):
            return f"{np.real(v):0}{np.imag(v):+}j"
        else:
            return str(v)

    @classmethod
    def from_record(cls, record: Union[et.a111.recording.Record, a121.Record]) -> "TableConverter":
        if isinstance(record, et.a111.recording.Record):
            return A111RecordTableConverter(record)
        elif isinstance(record, a121.Record):
            return A121RecordTableConverter(record)
        else:
            raise ValueError(f"Passed record ({record}) was of unexpected type.")


class A111RecordTableConverter(TableConverter):
    def __init__(self, record: et.a111.recording.Record) -> None:
        self._record = record

    def get_metadata_rows(self, sensor: int) -> list[np.ndarray]:
        depths = et.a111.get_range_depths(self._record.sensor_config, self._record.session_info)
        num_points = len(depths)
        rounded_depths = np.round(depths, decimals=6)

        if self._record.mode != et.a111.Mode.SPARSE:
            return rounded_depths
        else:
            spf = self._record.sensor_config.sweeps_per_frame
            sweep_numbers = np.repeat(range(spf), repeats=num_points).astype(int)
            depths_header = np.tile(rounded_depths, spf)
            return [sweep_numbers, depths_header]

    def convert(self, sensor: int) -> np.ndarray:
        """Converts data of a single sensor

        :param sensor: The sensor index
        :returns: 2D NDArray of cell values.
        """
        record = self._record
        sensor_index = sensor

        num_sensors = record.data.shape[1]
        if sensor_index >= num_sensors:
            raise ValueError(
                f"Invalid sensor index specified (index={sensor_index}). "
                f"Valid indices for this input file is one of {list(range(num_sensors))}"
            )

        data = record.data[:, sensor_index, :]
        dest_rows = []

        for x in data:
            row = np.ndarray.flatten(x)
            dest_rows.append([self.format_cell_value(v) for v in row])

        return np.array(dest_rows)

    def print_information(self, verbose: bool) -> None:
        config_dump = self.parse_config_dump(self._record.sensor_config_dump)
        print("=== Session info " + "=" * 43)
        for k, v in config_dump.items():
            print(f"{k:30} {v} ")
        print("=" * 60)
        print()

        if not verbose:
            return

        record = self._record
        print("Mode:", record.mode.name.lower())
        print()
        print(record.sensor_config)
        print()
        print("Session info")

        for k, v in record.session_info.items():
            print("  {:.<35} {}".format(k + " ", v))

        print()
        print("Data shape:", record.data.shape)
        print("Data dtype:", record.data.dtype)
        print()
        print("Last data info (first sensor):")

        for k, v in record.data_info[-1][0].items():
            print("  {:.<35} {}".format(k + " ", v))

        ts = record.sample_times
        if ts is not None and ts.size >= 2:
            print()
            mean_dt = (ts[-1] - ts[0]) / (ts.size - 1)
            mean_f = 1 / mean_dt
            print("Mean sample rate (client side): {:.2f} Hz".format(mean_f))

        print("\n")

        print("Module (processing) key:", record.module_key)

        if record.processing_config_dump is None:
            print("No processing config dump")
        else:
            print("Processing config dump")
            for k, v in json.loads(record.processing_config_dump).items():
                print("  {:.<35} {}".format(k + " ", v))

        print("\n")

        m = {
            "RSS version": record.rss_version,
            "acconeer.exptool library version": record.lib_version,
            "Timestamp": record.timestamp,
        }

        for k, v in m.items():
            print("{:.<37} {}".format(k + " ", v))

        if record.note:
            print()
            print("Note: " + str(record.note))

    @staticmethod
    def parse_config_dump(config: str):
        context = {"null": None, "true": True, "false": False}
        return eval(config, context)


class A121RecordTableConverter(TableConverter):
    def __init__(self, record: a121.Record) -> None:
        self._record = record

    def _results_of_sensor_id(self, sensor_id: int) -> list[a121.Result]:
        return [
            ext_result.get(sensor_id)
            for ext_result_group in self._record.extended_results
            for ext_result in ext_result_group
            if ext_result.get(sensor_id) is not None
        ]

    def _unique_sensor_configs_of_sensor_id(self, sensor_id: int) -> list[a121.SensorConfig]:
        # FIXME: this is not a `set` as SensorConfig is not hash-able
        sensor_configs = [
            sensor_config
            for _, sid, sensor_config in _core.utils.iterate_extended_structure(
                self._record.session_config.groups
            )
            if sid == sensor_id
        ]
        unique_sensor_configs = []
        for sensor_config in sensor_configs:
            if sensor_config not in unique_sensor_configs:
                unique_sensor_configs.append(sensor_config)
        return unique_sensor_configs

    def _convert_single_sensor_config(
        self,
        sensor: int,
    ) -> np.ndarray:
        """This function handles the case where the sensor at "sensor ID" is configured
        with a single unique `SensorConfig`, possibly in multiple groups.

        :param sensor: The sensor ID.
        :returns: 2D NDArray of cell values.
        """
        return np.array(
            [
                [self.format_cell_value(v) for v in result.frame.flatten()]
                for result in self._results_of_sensor_id(sensor)
            ]
        )

    def _convert_multiple_sensor_config(self, sensor: int) -> np.ndarray:
        """
        This function handles the case where the sensor at "sensor ID" is configured
        with a multiple unique `SensorConfig` across groups.

        :param sensor: The sensor ID.
        :returns: 2D NDArray of cell values.
        """
        raise NotImplementedError(
            "This record contains data where a single sensor has multiple configuration through\n"
            + "the use of groups. Exporting this kind of record is not possible at the moment.\n"
            + "\n"
            + "If this is a feature that you are interested in, please get in contact with us!"
        )

    def convert(self, sensor: int) -> np.ndarray:
        """Converts data of a single sensor

        :param sensor: The sensor index
        :returns: 2D NDArray of cell values.
        """
        unique_sensor_configs = self._unique_sensor_configs_of_sensor_id(sensor)

        if len(unique_sensor_configs) == 1:
            return self._convert_single_sensor_config(sensor)

        if len(unique_sensor_configs) > 1:
            return self._convert_multiple_sensor_config(sensor)

        raise ValueError(f"This record contains no data of sensor with id = {sensor}")

    def _get_metadata_rows_single_sensor_config(self, sensor: int) -> list[np.ndarray]:
        (sensor_config,) = self._unique_sensor_configs_of_sensor_id(sensor)
        (metadata,) = (
            meta
            for _, sid, meta in _core.utils.iterate_extended_structure(
                self._record.extended_metadata
            )
            if sid == sensor
        )
        depths, _ = algo.get_distances_m(sensor_config, metadata)

        sweep_numbers = np.repeat(
            range(sensor_config.sweeps_per_frame), repeats=sensor_config.num_points
        ).astype(int)
        depths_header = np.tile(depths, sensor_config.sweeps_per_frame)
        return [sweep_numbers, depths_header]

    def _get_metadata_rows_multiple_sensor_config(self, sensor: int) -> list[np.ndarray]:
        raise NotImplementedError(
            "This record contains data where a single sensor has multiple configuration through\n"
            + "the use of groups. Exporting this kind of record is not possible at the moment.\n"
            + "\n"
            + "If this is a feature that you are interested in, please get in contact with us!"
        )

    def get_metadata_rows(self, sensor: int) -> list[np.ndarray]:
        unique_sensor_configs = self._unique_sensor_configs_of_sensor_id(sensor)

        if len(unique_sensor_configs) == 1:
            return self._get_metadata_rows_single_sensor_config(sensor)
        else:
            return self._get_metadata_rows_multiple_sensor_config(sensor)

    def print_information(self, verbose: bool = False) -> None:
        extended = self._record.session_config.extended

        print("=== Session config " + "=" * 41)
        print(self._record.session_config)

        if not verbose:
            print("=" * 60)
            return

        print("=== Meta data " + "=" * 46)
        pprint(self._record.extended_metadata if extended else self._record.metadata)
        print("=== Server info " + "=" * 44)
        pprint(self._record.server_info)
        print("=== Client info " + "=" * 44)
        pprint(self._record.client_info)
        print("=== Misc. " + "=" * 50)
        print(f"Exptool version:  {self._record.lib_version}")
        print(f"Number of frames: {self._record.num_frames}")
        print(f"Timestamp:        {self._record.timestamp}")
        print(f"UUID:             {self._record.uuid}")

        print("=" * 60)


def _check_files(input_file, output_file, force):
    if not os.path.exists(input_file):
        print(f'The input file ("{input_file}") can not be found.')
        exit(1)

    if os.path.exists(output_file) and not force:
        print(f'The output file ("{output_file}") already exists.')
        print('If you know what you are doing; overwrite it with "-f".')
        exit(1)


def load_file(input_file: str) -> tuple[Union[et.a111.recording.Record, a121.Record], str]:
    for loader, generation in [(a121.load_record, "a121"), (et.a111.recording.load, "a111")]:
        try:
            return loader(input_file), generation
        except Exception:
            pass

    raise Exception("The specified file was neither A111 or A121. Cannot load.")


def get_default_sensor_id_or_index(namespace: argparse.Namespace, generation: str) -> int:
    try:
        return namespace.sensor
    except AttributeError:
        return 1 if generation == "a121" else 0


def main():
    parser = ConvertToCsvArgumentParser()
    args = parser.parse_args()

    # Convert to real delimiter given to csv module
    delimiter = {"c": ",", "t": "\t"}.get(args.delimiter)

    _check_files(args.input_file, args.output_file, args.force)
    print(f"Reading from {args.input_file!r} ... \n")
    record, generation = load_file(args.input_file)
    sensor = get_default_sensor_id_or_index(args, generation)

    table_converter = TableConverter.from_record(record)

    try:
        data_table = table_converter.convert(sensor=sensor)
    except Exception as e:
        print(e)
        exit(1)

    table_converter.print_information(verbose=args.verbose)
    print()

    if args.sweep_as_column:
        data_table = data_table.T

    with open(args.output_file, "w") as f:
        writer = csv.writer(f, delimiter=delimiter)

        if args.add_sweep_metadata:
            metadata_rows = table_converter.get_metadata_rows(sensor=sensor)
            print(f"Writing {len(metadata_rows)} rows of metadata ...")
            for row in metadata_rows:
                writer.writerow(row)

        print(f"Writing data with shape {data_table.shape} ...")
        for row in data_table:
            writer.writerow(row)

    print("Success!")


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import os
import sys

import acconeer.exptool as et


def main():
    args = et.a111.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    filename = "data.h5"
    if os.path.exists(filename):
        print("File '{}' already exists, won't overwrite".format(filename))
        sys.exit(1)

    client = et.a111.Client(**et.a111.get_client_args(args))

    config = et.a111.EnvelopeServiceConfig()
    config.sensor = args.sensors
    config.update_rate = 30

    session_info = client.setup_session(config)

    recorder = et.a111.recording.Recorder(sensor_config=config, session_info=session_info)

    client.start_session()

    n = 100
    for i in range(n):
        data_info, data = client.get_next()
        recorder.sample(data_info, data)
        print("Sampled {:>4}/{}".format(i + 1, n), end="\r", flush=True)

    print()

    client.disconnect()

    record = recorder.close()
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    et.a111.recording.save(filename, record)
    print("Saved to '{}'".format(filename))


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import configparser
import fnmatch
import subprocess
import sys


def main():
    config = configparser.ConfigParser()
    config.read("setup.cfg")
    section = "check_line_length"

    line_length = int(config.get(section, "line_length"))
    include = config.get(section, "include", fallback="")
    include = [s.strip() for s in include.split(",")]
    include = [s for s in include if s]

    p = subprocess.run(
        ["git", "ls-files"],
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
        check=True,
    )

    filenames = p.stdout.decode().splitlines()

    failed = False
    for filename in filenames:
        if not any(fnmatch.fnmatch(filename, p) for p in include):
            continue

        try:
            check(filename, line_length)
        except UnicodeDecodeError:
            pass
        except Exception as e:
            failed = True
            print("{}: {}".format(filename, e))

    return failed


def check(filename, max_line_length):
    with open(filename, "r") as f:
        lines = f.read().splitlines()

    for i, line in enumerate(lines):
        n = len(line)
        if n > max_line_length:
            msg = f"Line {i + 1} too long ({n} > {max_line_length} characters)"
            raise Exception(msg)


if __name__ == "__main__":
    sys.exit(int(main()))



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import multiprocessing as mp
import queue
import signal
from time import sleep, time


class PGProcess:
    def __init__(self, updater, max_freq=60):
        self._queue = mp.Queue()
        self._exit_event = mp.Event()

        args = (
            self._queue,
            self._exit_event,
            updater,
            max_freq,
        )

        self._process = mp.Process(target=pg_process_program, args=args, daemon=True)

    def start(self):
        self._process.start()

    def put_data(self, data):
        if self._exit_event.is_set() or self._process.exitcode is not None:
            self.close()
            raise PGProccessDiedException

        try:
            self._queue.put(data)
        except BrokenPipeError:
            self.close()
            raise PGProccessDiedException

    def close(self):
        if self._process.exitcode is None:
            self._exit_event.set()
            self._process.join(1)

        if self._process.exitcode is None:
            self._process.terminate()
            self._process.join(1)

        if self._process.exitcode is None:
            raise RuntimeError


def pg_process_program(q, exit_event, updater, max_freq):
    from PySide6 import QtWidgets

    import pyqtgraph as pg

    signal.signal(signal.SIGINT, signal.SIG_IGN)

    app = QtWidgets.QApplication([])
    pg.setConfigOption("background", "w")
    pg.setConfigOption("foreground", "k")
    pg.setConfigOptions(antialias=True)
    win = pg.GraphicsLayoutWidget()
    win.closeEvent = lambda _: exit_event.set()

    updater.setup(win.ci)

    win.show()
    app.processEvents()

    while not exit_event.is_set():
        data = None
        try:
            data = q.get(timeout=0.1)
            while True:
                data = q.get(timeout=0.001)
        except queue.Empty:
            pass

        data_time = time()

        if data is not None:
            updater.update(data)

        app.processEvents()

        if max_freq and data is not None:
            sleep_time = 1 / max_freq - (time() - data_time)
            if sleep_time > 0.005:
                sleep(sleep_time)

    win.close()
    app.closeAllWindows()

    try:
        while True:
            q.get(timeout=0.001)
    except Exception:
        pass


class ExamplePGUpdater:
    def __init__(self):
        pass

    def setup(self, win):
        import pyqtgraph as pg

        win.setWindowTitle("PG updater example")
        self.plot = win.addPlot()
        self.curve = self.plot.plot(pen=pg.mkPen("k", width=3))

    def update(self, data):
        x, y = data
        self.curve.setData(x, y)


class PGProccessDiedException(Exception):
    pass


# Example usage:
if __name__ == "__main__":
    import numpy as np

    pg_updater = ExamplePGUpdater()
    pg_process = PGProcess(pg_updater)
    pg_process.start()

    x = np.linspace(0, 10, 100)
    t0 = time()
    t = 0
    while t < 10:
        t = time() - t0
        y = np.sin(x + t)

        try:
            pg_process.put_data([x, y])
        except PGProccessDiedException:
            exit()

        sleep(0.01)

    pg_process.close()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from . import pidgets
from .attrs_config_editor import AttrsConfigEditor
from .metadata_view import ExtendedMetadataView, MetadataView, SmartMetadataView
from .perf_calc_view import ExtendedPerfCalcView, PerfCalcView, SmartPerfCalcView
from .session_config_editor import SessionConfigEditor
from .types import PidgetFactoryMapping
from .utils import GridGroupBox, HorizontalGroupBox, VerticalGroupBox



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import numpy as np

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.virtual_button import (
    Processor,
    ProcessorConfig,
    ProcessorResult,
    get_sensor_config,
)


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client(**a121.get_client_args(args))
    client.connect()

    processor_config = ProcessorConfig()

    sensor_config = get_sensor_config()

    metadata = client.setup_session(sensor_config)
    client.start_session()

    processor = Processor(
        sensor_config=sensor_config,
        metadata=metadata,
        processor_config=processor_config,
    )

    pg_updater = PGUpdater()
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        result = client.get_next()
        processor_result = processor.process(result)
        try:
            pg_process.put_data(processor_result)
        except et.PGProccessDiedException:
            break

    print("Disconnecting...")
    pg_process.close()
    client.stop_session()
    client.disconnect()


class PGUpdater:
    def __init__(self):
        self.detection_history = np.full((2, 100), np.NaN)

    def setup(self, win):

        close_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("Close detection")
        )
        far_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("Far detection")
        )
        self.close_text_item = pg.TextItem(
            html=close_html,
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.far_text_item = pg.TextItem(
            html=far_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )

        self.detection_history_plot = win.addPlot(row=1, col=2)
        self.detection_history_plot.setMenuEnabled(False)
        self.detection_history_plot.setMouseEnabled(x=False, y=False)
        self.detection_history_plot.hideButtons()
        self.detection_history_plot.showGrid(x=True, y=True)
        self.detection_history_plot.setYRange(-0.1, 1.8)
        self.detection_history_curve_close = self.detection_history_plot.plot(
            pen=et.utils.pg_pen_cycler(1, width=5)
        )
        self.detection_history_curve_far = self.detection_history_plot.plot(
            pen=et.utils.pg_pen_cycler(0, width=5)
        )

        pos_top = (100 / 2, 1.8)
        pos_bottom = (100 / 2, 1.5)
        self.close_text_item.setPos(*pos_bottom)
        self.far_text_item.setPos(*pos_top)
        self.detection_history_plot.addItem(self.close_text_item)
        self.detection_history_plot.addItem(self.far_text_item)
        self.close_text_item.hide()
        self.far_text_item.hide()

    def update(self, result: ProcessorResult):
        detection = np.array([result.detection_close, result.detection_far])
        self.detection_history = np.roll(self.detection_history, -1, axis=1)
        self.detection_history[:, -1] = detection

        self.detection_history_curve_close.setData(self.detection_history[0])
        self.detection_history_curve_far.setData(self.detection_history[1])
        print(detection)

        if detection[0]:
            self.close_text_item.show()
        else:
            self.close_text_item.hide()

        if detection[1]:
            self.far_text_item.show()
        else:
            self.far_text_item.hide()


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import acconeer.exptool as et


def main():
    args = et.a111.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = et.a111.Client(**et.a111.get_client_args(args))

    config = et.a111.IQServiceConfig()
    config.sensor = args.sensors
    config.range_interval = [0.2, 0.6]
    config.update_rate = 50

    info = client.start_session(config)

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    fc = et.utils.FreqCounter(num_bits=(4 * 8 * info["data_length"]))

    while not interrupt_handler.got_signal:
        info, data = client.get_next()
        fc.tick()

    print("\nDisconnecting...")
    client.disconnect()


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from enum import Enum, unique
from typing import Optional, Type, TypeVar


T = TypeVar("T", bound=Enum)


def find_by_value(enum: Type[T], value: object) -> Optional[T]:
    for member in enum:
        if member.value == value:
            return member
    return None


def find_by_name(enum: Type[T], name: object) -> Optional[T]:
    for member in enum:
        if member.name == name:
            return member
    return None


def find_by_lowercase_name(enum: Type[T], lowercase_name: object) -> Optional[T]:
    for member in enum:
        if member.name.lower() == lowercase_name:
            return member
    return None


@unique
class Profile(Enum):
    """Profile"""

    PROFILE_1 = 1
    PROFILE_2 = 2
    PROFILE_3 = 3
    PROFILE_4 = 4
    PROFILE_5 = 5

    @classmethod
    def _missing_(cls, value: object) -> Optional[Profile]:
        return find_by_value(cls, value) or find_by_name(cls, value)


@unique
class PRF(Enum):
    """Pulse Repetition Frequency (PRF)

    Pulse Repetition Frequency, PRF, is the frequency at which pulses are sent out from the radar
    system. The measurement time is approximately proportional to the PRF. The higher the PRF, the
    shorter the measurement time.

    This parameter sets the Maximum Measurable Distance, MMD, that can be achieved. MMD is the
    maximum value for the end point, i.e., the start point + (number of points * step length). For
    example, an MMD of 7.0 m means that the range cannot be set further out than 7.0 m.

    It also sets the Maximum Unambiguous Range, MUR, that can be achieved. MUR is the maximum
    distance at which an object can be located to guarantee that its reflection corresponds to the
    most recent transmitted pulse. Objects farther away than the MUR may fold into the measured
    range. For example, with a MUR of 11.5 m, an object at 13.5 m could become visible at 2 m.

    ================= ======== ====== ======
    PRF Setting            PRF    MMD    MUR
    ================= ======== ====== ======
    PRF_19_5_MHZ [*]_ 19.5 MHz  3.1 m  7.7 m
    PRF_13_0_MHZ      13.0 MHz  7.0 m 11.5 m
    PRF_8_7_MHZ        8.7 MHz 12.7 m 17.3 m
    PRF_6_5_MHZ        6.5 MHz 18.5 m 23.1 m
    ================= ======== ====== ======

    .. [*] 19.5MHz is only available for profile 1 and 2.
    """

    PRF_19_5_MHz = 19500000
    PRF_13_0_MHz = 13000000
    PRF_8_7_MHz = 8700000
    PRF_6_5_MHz = 6500000

    @property
    def frequency(self) -> int:
        return self.value

    @classmethod
    def _missing_(cls, value: object) -> Optional[PRF]:
        return find_by_value(cls, value) or find_by_name(cls, value)


@unique
class IdleState(Enum):
    """Idle state

    Idle state :attr:`DEEP_SLEEP` is the deepest state where as much of the
    sensor hardware as possible is shut down and idle state :attr:`READY` is
    the lightest state where most of the sensor hardware is kept on.

    :attr:`DEEP_SLEEP` is the slowest to transition from while :attr:`READY` is
    the fastest.
    """

    DEEP_SLEEP = 0
    SLEEP = 1
    READY = 2

    @classmethod
    def _missing_(cls, value: object) -> Optional[IdleState]:
        if find_by_value(cls, value) is not None:
            return find_by_value(cls, value)

        if find_by_name(cls, value) is not None:
            return find_by_name(cls, value)

        if find_by_lowercase_name(cls, value) is not None:
            return find_by_lowercase_name(cls, value)

        return None

    def is_deeper_than(self, other: IdleState) -> bool:
        return self.value < other.value



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from functools import partial
from typing import Any, Generic, Mapping, Optional, TypeVar

import attrs

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QVBoxLayout, QWidget

from acconeer.exptool.a121._core import Criticality

from .pidgets import ParameterWidget
from .types import PidgetFactoryMapping
from .utils import VerticalGroupBox


T = TypeVar("T")


class AttrsConfigEditor(QWidget, Generic[T]):
    _config: Optional[T]

    sig_update = Signal(object)

    def __init__(
        self, title: str, factory_mapping: PidgetFactoryMapping, parent: Optional[QWidget] = None
    ) -> None:
        super().__init__(parent=parent)
        self._config = None
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(11)
        group_box = VerticalGroupBox(title, parent=self)
        self.layout().addWidget(group_box)

        self._pidget_mapping: Mapping[str, ParameterWidget] = {}

        for aspect, factory in factory_mapping.items():
            pidget = factory.create(group_box)
            pidget.sig_parameter_changed.connect(partial(self._update_config_aspect, aspect))
            group_box.layout().addWidget(pidget)

            self._pidget_mapping[aspect] = pidget

    def set_data(self, config: Optional[T]) -> None:
        self._config = config

    def sync(self) -> None:
        self._update_pidgets()

    def setEnabled(self, enabled: bool) -> None:
        super().setEnabled(enabled and self._config is not None)

    def _broadcast(self) -> None:
        self.sig_update.emit(self._config)

    def _update_pidgets(self) -> None:
        if self._config is None:
            return

        for aspect, pidget in self._pidget_mapping.items():
            config_value = getattr(self._config, aspect)
            pidget.set_parameter(config_value)

    def _update_config_aspect(self, aspect: str, value: Any) -> None:
        if self._config is None:
            return

        try:
            self._config = attrs.evolve(self._config, **{aspect: value})
        except Exception as e:
            self._pidget_mapping[aspect].set_note_text(e.args[0], Criticality.ERROR)
        else:
            self._pidget_mapping[aspect].set_note_text(None)

        self._broadcast()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import logging
from typing import Tuple, Type

import numpy as np
import numpy.typing as npt

from PySide6.QtGui import QTransform

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121 import algo
from acconeer.exptool.a121.algo._plugins import (
    ProcessorBackendPluginBase,
    ProcessorPlotPluginBase,
    ProcessorViewPluginBase,
)
from acconeer.exptool.app.new import AppModel, Plugin, PluginFamily, PluginGeneration
from acconeer.exptool.app.new.ui.plugin import PidgetFactoryMapping, pidgets

from ._processor import (
    AmplitudeMethod,
    Processor,
    ProcessorConfig,
    ProcessorResult,
    get_sensor_config,
)


log = logging.getLogger(__name__)


class BackendPlugin(ProcessorBackendPluginBase[ProcessorConfig, ProcessorResult]):
    @classmethod
    def get_processor_cls(cls) -> Type[Processor]:
        return Processor

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def get_default_sensor_config(cls) -> a121.SensorConfig:
        return get_sensor_config()


class ViewPlugin(ProcessorViewPluginBase[ProcessorConfig]):
    @classmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        return {
            "amplitude_method": pidgets.EnumParameterWidgetFactory(
                enum_type=AmplitudeMethod,
                name_label_text="Amplitude method:",
                label_mapping={
                    AmplitudeMethod.COHERENT: "Coherent",
                    AmplitudeMethod.NONCOHERENT: "Non-coherent",
                    AmplitudeMethod.FFT_MAX: "FFT Max",
                },
            )
        }

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig


class PlotPlugin(ProcessorPlotPluginBase[ProcessorResult]):
    def __init__(self, *, plot_layout: pg.GraphicsLayout, app_model: AppModel) -> None:
        super().__init__(plot_layout=plot_layout, app_model=app_model)
        self.smooth_max = et.utils.SmoothMax()

    def setup(self, metadata: a121.Metadata, sensor_config: a121.SensorConfig) -> None:
        self.distances_m, step_length_m = algo.get_distances_m(sensor_config, metadata)
        vels, vel_res = algo.get_approx_fft_vels(sensor_config)

        self.ampl_plot = self._create_amplitude_plot(self.plot_layout)
        self.ampl_curve = self._create_amplitude_curve(0, self.distances_m)
        self.ampl_plot.addDataItem(self.ampl_curve)

        self.plot_layout.nextRow()

        self.phase_plot = self._create_phase_plot(self.plot_layout)
        self.phase_curve = self._create_phase_curve(0)
        self.phase_plot.addDataItem(self.phase_curve)

        self.plot_layout.nextRow()

        self.ft_plot, self.ft_im = self._create_fft_plot(
            self.plot_layout,
            distances_m=self.distances_m,
            step_length_m=step_length_m,
            vels=vels,
            vel_res=vel_res,
        )

    def update(self, processor_result: ProcessorResult) -> None:
        ampls = processor_result.amplitudes
        self.ampl_curve.setData(self.distances_m, ampls)
        self.phase_curve.setData(self.distances_m, processor_result.phases)
        self.ampl_plot.setYRange(0, self.smooth_max.update(ampls))
        dvm = processor_result.distance_velocity_map
        self.ft_im.updateImage(
            dvm.T,
            levels=(0, 1.05 * np.max(dvm)),
        )

    @staticmethod
    def _create_amplitude_curve(
        cycle_num: int, depths_m: npt.NDArray[np.float_]
    ) -> pg.PlotDataItem:
        pen = et.utils.pg_pen_cycler(cycle_num)

        if len(depths_m) > 32:
            return pg.PlotDataItem(pen=pen)
        else:
            brush = et.utils.pg_brush_cycler(cycle_num)
            return pg.PlotDataItem(
                pen=pen, symbol="o", symbolSize=5, symbolBrush=brush, symbolPen="k"
            )

    @staticmethod
    def _create_phase_curve(cycle_num: int) -> pg.PlotDataItem:
        brush = et.utils.pg_brush_cycler(cycle_num)
        return pg.PlotDataItem(
            pen=None, symbol="o", symbolSize=5, symbolBrush=brush, symbolPen="k"
        )

    @staticmethod
    def _create_amplitude_plot(parent: pg.GraphicsLayout) -> pg.PlotItem:
        ampl_plot = parent.addPlot()
        ampl_plot.setMenuEnabled(False)
        ampl_plot.showGrid(x=True, y=True)
        ampl_plot.setLabel("left", "Amplitude")
        return ampl_plot

    @staticmethod
    def _create_phase_plot(parent: pg.GraphicsLayout) -> pg.PlotItem:
        phase_plot = parent.addPlot()
        phase_plot.setMenuEnabled(False)
        phase_plot.showGrid(x=True, y=True)
        phase_plot.setLabel("left", "Phase")
        phase_plot.setYRange(-np.pi, np.pi)
        phase_plot.getAxis("left").setTicks(et.utils.pg_phase_ticks)
        return phase_plot

    @staticmethod
    def _create_fft_plot(
        parent: pg.GraphicsLayout,
        *,
        distances_m: npt.NDArray[np.float_],
        step_length_m: float,
        vels: npt.NDArray[np.float_],
        vel_res: float,
    ) -> Tuple[pg.PlotItem, pg.ImageItem]:
        transform = QTransform()
        transform.translate(distances_m[0], vels[0] - 0.5 * vel_res)
        transform.scale(step_length_m, vel_res)

        im = pg.ImageItem(autoDownsample=True)
        im.setLookupTable(et.utils.pg_mpl_cmap("viridis"))
        im.setTransform(transform)

        plot = parent.addPlot()
        plot.setMenuEnabled(False)
        plot.setLabel("bottom", "Distance (m)")
        plot.setLabel("left", "Velocity (m/s)")
        plot.addItem(im)

        return plot, im


SPARSE_IQ_PLUGIN = Plugin(
    generation=PluginGeneration.A121,
    key="sparse_iq",
    title="Sparse IQ",
    description="Basic usage of the sparse IQ service.",
    family=PluginFamily.SERVICE,
    backend_plugin=BackendPlugin,
    plot_plugin=PlotPlugin,
    view_plugin=ViewPlugin,
)
